import numpy as np

def gumbel_fd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,i] = (1/v2 - np.exp(-(-v1 + x)/v2)/v2)*np.exp(-np.exp(-(-v1 + x)/v2) - (-v1 + x)/v2)/v2 
		result[1,i] = ((-v1 + x)/v2**2 - (-v1 + x)*np.exp(-(-v1 + x)/v2)/v2**2)*np.exp(-np.exp(-(-v1 + x)/v2) - (-v1 + x)/v2)/v2 - np.exp(-np.exp(-(-v1 + x)/v2) - (-v1 + x)/v2)/v2**2 
	if len(x_all) == 1:
		result = result[:,0]
	return result 

def gumbel_fdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = ((np.exp((v1 - x)/v2) - 1)**2 - np.exp((v1 - x)/v2))*np.exp(-np.exp(-(-v1 + x)/v2) - (-v1 + x)/v2)/v2**3
		result[0,1,i] = (2*np.exp((v1 - x)/v2) - 2 - (v1 - x)*(np.exp((v1 - x)/v2) - 1)**2/v2 + (v1 - x)*np.exp((v1 - x)/v2)/v2)*np.exp(-np.exp(-(-v1 + x)/v2) - (-v1 + x)/v2)/v2**3
		result[1,0,i] = (2*np.exp((v1 - x)/v2) - 2 - (v1 - x)*(np.exp((v1 - x)/v2) - 1)**2/v2 + (v1 - x)*np.exp((v1 - x)/v2)/v2)*np.exp(-np.exp(-(-v1 + x)/v2) - (-v1 + x)/v2)/v2**3
		result[1,1,i] = (2 - 2*(v1 - x)*(np.exp((v1 - x)/v2) - 1)/v2 - (v1 - x)*(2*np.exp((v1 - x)/v2) - 2 - (v1 - x)*(np.exp((v1 - x)/v2) - 1)**2/v2 + (v1 - x)*np.exp((v1 - x)/v2)/v2)/v2)*np.exp(-np.exp(-(-v1 + x)/v2) - (-v1 + x)/v2)/v2**3
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def gumbel_pd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,i] = -np.exp(-(-v1 + x)/v2)*np.exp(-np.exp(-(-v1 + x)/v2))/v2 
		result[1,i] = -(-v1 + x)*np.exp(-(-v1 + x)/v2)*np.exp(-np.exp(-(-v1 + x)/v2))/v2**2 
	if len(x_all) == 1:
		result = result[:,0]
	return result 

def gumbel_pdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = (np.exp((v1 - x)/v2) - 1)*np.exp((v1 - x)/v2)*np.exp(-np.exp((v1 - x)/v2))/v2**2
		result[0,1,i] = (1 - (v1 - x)*np.exp((v1 - x)/v2)/v2 + (v1 - x)/v2)*np.exp((v1 - x)/v2)*np.exp(-np.exp((v1 - x)/v2))/v2**2
		result[1,0,i] = (1 - (v1 - x)*np.exp((v1 - x)/v2)/v2 + (v1 - x)/v2)*np.exp((v1 - x)/v2)*np.exp(-np.exp((v1 - x)/v2))/v2**2
		result[1,1,i] = (v1 - x)*(-2 + (v1 - x)*np.exp((v1 - x)/v2)/v2 - (v1 - x)/v2)*np.exp((v1 - x)/v2)*np.exp(-np.exp((v1 - x)/v2))/v2**3
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def gumbel_logfdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = -np.exp((v1 - x)/v2)/v2**2
		result[0,1,i] = (np.exp((v1 - x)/v2) - 1 + (v1 - x)*np.exp((v1 - x)/v2)/v2)/v2**2
		result[1,0,i] = (np.exp((v1 - x)/v2) - 1 + (v1 - x)*np.exp((v1 - x)/v2)/v2)/v2**2
		result[1,1,i] = (1 - 2*(v1 - x)*np.exp((v1 - x)/v2)/v2 + 2*(v1 - x)/v2 - (v1 - x)**2*np.exp((v1 - x)/v2)/v2**2)/v2**2
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def gumbel_logfddd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, dim, dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,0,i] = -np.exp((v1 - x)/v2)/v2**3 
		result[0,0,1,i] = (2 + (v1 - x)/v2)*np.exp((v1 - x)/v2)/v2**3 
		result[0,1,0,i] = (2 + (v1 - x)/v2)*np.exp((v1 - x)/v2)/v2**3 
		result[0,1,1,i] = (-2*np.exp((v1 - x)/v2) + 2 - 4*(v1 - x)*np.exp((v1 - x)/v2)/v2 - (v1 - x)**2*np.exp((v1 - x)/v2)/v2**2)/v2**3 
		result[1,0,0,i] = (2 + (v1 - x)/v2)*np.exp((v1 - x)/v2)/v2**3 
		result[1,0,1,i] = (-2*np.exp((v1 - x)/v2) + 2 - 4*(v1 - x)*np.exp((v1 - x)/v2)/v2 - (v1 - x)**2*np.exp((v1 - x)/v2)/v2**2)/v2**3 
		result[1,1,0,i] = (-2*np.exp((v1 - x)/v2) + 2 - 4*(v1 - x)*np.exp((v1 - x)/v2)/v2 - (v1 - x)**2*np.exp((v1 - x)/v2)/v2**2)/v2**3 
		result[1,1,1,i] = (-2 + 6*(v1 - x)*np.exp((v1 - x)/v2)/v2 - 6*(v1 - x)/v2 + 6*(v1 - x)**2*np.exp((v1 - x)/v2)/v2**2 + (v1 - x)**3*np.exp((v1 - x)/v2)/v2**3)/v2**3 
	if len(x_all) == 1:
		result = result[:,:,:,0]
	return result 



import scipy
import sympy
import numpy as np
import autodiff

x, v1, v2, v3 = sympy.symbols('x v1 v2 v3')

f_gev_density = (1/v2)*((1+v3*((x-v1)/v2))**(-1/v3-1))*sympy.exp(-((1+v3*(x-v1)/v2)**(-1/v3)))
f_gev_cdf = sympy.exp(-((1+v3*(x-v1)/v2)**(-1/v3)))
f_gev_log_density = -sympy.log(v2)-(1+1/v3)*sympy.log(1+v3*(x-v1)/v2)-(1+v3*(x-v1)/v2)**(-1/v3)
                                 
with open('gev_autodiff_python_ver.py', 'w') as f:
    f.write('import numpy as np\n')
    f.write(autodiff.diff_to_code(f_gev_density, [v1, v2, v3], 'gev_fd', 1))
    f.write(autodiff.diff_to_code(f_gev_density, [v1, v2, v3], 'gev_fdd', 2))
    f.write(autodiff.diff_to_code(f_gev_cdf, [v1, v2, v3], 'gev_pd', 1))
    f.write(autodiff.diff_to_code(f_gev_cdf, [v1, v2, v3], 'gev_pdd', 2))
    f.write(autodiff.diff_to_code(f_gev_log_density, [v1, v2, v3], 'gev_logfd', 1))
    f.write(autodiff.diff_to_code(f_gev_log_density, [v1, v2, v3], 'gev_logfdd', 2))
    f.write(autodiff.diff_to_code(f_gev_log_density, [v1, v2, v3], 'gev_logfddd', 3))

f.close()


import sympy
import autodiff

x, t1, t2, v1, v2, v3, v4, v5 = sympy.symbols('x t1 t2 v1 v2 v3 v4 v5')

f = (sympy.exp(-v3-v4*t2))*((1+v5*((x-v1-v2*t1)*sympy.exp(-v3-v4*t2)))**(-1/v5-1))*sympy.exp(-((1+v5*(x-v1-v2*t1)*sympy.exp(-v3-v4*t2))**(-1/v5)))
p = sympy.exp(-((1+v5*(x-v1-v2*t1)*sympy.exp(-v3-v4*t2))**(-1/v5)))
logf = -v3-v4*t2-(1+1/v5)*sympy.log(1+v5*(x-v1-v2*t1)*sympy.exp(-v3-v4*t2))-(1+v5*(x-v1-v2*t1)*sympy.exp(-v3-v4*t2))**(-1/v5)
                                 
with open('output_gev_p12.py', 'w') as file:
    file.write('import numpy as np\n')
    autodiff.diff_to_file(file, f, [v1, v2, v3, v4, v5], 'gev_p12_fd', 1, xt=True)
    autodiff.diff_to_file(file, f, [v1, v2, v3, v4, v5], 'gev_p12_fdd', 2, xt=True)
    autodiff.diff_to_file(file, p, [v1, v2, v3, v4, v5], 'gev_p12_pd', 1, xt=True)
    autodiff.diff_to_file(file, p, [v1, v2, v3, v4, v5], 'gev_p12_pdd', 2, xt=True)
    autodiff.diff_to_file(file, logf, [v1, v2, v3, v4, v5], 'gev_p12_logfdd', 2, xt=True)
    autodiff.diff_to_file(file, logf, [v1, v2, v3, v4, v5], 'gev_p12_logfddd', 3, xt=True)

file.close()


import numpy as np
import scipy.stats as stats
import scipy.optimize as optimize
import scipy.integrate as integrate
import sys

from . import utils as cp_utils
from . import evaluate_dmgs_equation as cp_dmgs
from . import gamma_derivs


def gamma_waic(waicscores, x, v1hat, fd1, v2hat, fd2, lddi, lddd, lambdad, aderivs):
    """
    Waic
    
    Parameters
    ----------
    waicscores : bool
        Whether to calculate WAIC scores
    x : array-like
        Data values
    v1hat : float
        Shape parameter estimate
    fd1 : float
        Step size for v1 finite differences
    v2hat : float
        Scale parameter estimate  
    fd2 : float
        Step size for v2 finite differences
    lddi : ndarray
        Inverse of second derivative of observed log-likelihood
    lddd : ndarray
        Third derivative of observed log-likelihood
    lambdad : array-like
        Lambda parameter vector
    aderivs : bool
        Whether to use analytical derivatives
        
    Returns
    -------
    dict
        Dictionary with waic1 and waic2 results
    """
    if waicscores:
        if aderivs:
            f1f = gamma_derivs.gamma_f1fa(x, v1hat, v2hat)
        else:
            f1f = gamma_f1f(x, v1hat, fd1, v2hat, fd2)
        
        if aderivs:
            f2f = gamma_derivs.gamma_f2fa(x, v1hat, v2hat)
        else:
            f2f = gamma_f2f(x, v1hat, fd1, v2hat, fd2)
        
        fhatx = stats.gamma.pdf(x, a=v1hat, scale=v2hat)
        waic = cp_utils.make_waic(x, fhatx, lddi, lddd, f1f, lambdad, f2f, dim=2)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
    else:
        waic1 = "waicscores not selected"
        waic2 = "waicscores not selected"
    
    return {'waic1': waic1, 'waic2': waic2}

def gamma_logf(params, x):
    """
    Logf for RUST
    
    Parameters
    ----------
    params : array-like
        Parameters [shape, scale]
    x : array-like
        Data values
        
    Returns
    -------
    float
        Log-likelihood value
    """
    sh = max(params[0], sys.float_info.epsilon)
    sc = max(params[1], sys.float_info.epsilon)
    logf = np.sum(stats.gamma.logpdf(x, a=sh, scale=sc)) - np.log(sh) - np.log(sc)
    return logf

def gamma_loglik(vv, x):
    """
    log-likelihood function
    
    Parameters
    ----------
    vv : array-like
        Parameters [shape, scale]
    x : array-like
        Data values
        
    Returns
    -------
    float
        Log-likelihood value
    """
    n = len(x)
    loglik = np.sum(stats.gamma.logpdf(x, a=max(vv[0], sys.float_info.epsilon), 
                                      scale=max(vv[1], sys.float_info.epsilon)))
    return loglik

def gamma_lmn(x, v1, fd1, v2, fd2, mm, nn):
    """
    One component of the second derivative of the normalized log-likelihood
    
    Parameters
    ----------
    x : array-like
        Data values
    v1 : float
        Shape parameter
    fd1 : float
        Step size for v1 finite differences
    v2 : float
        Scale parameter
    fd2 : float
        Step size for v2 finite differences
    mm : int
        First parameter index (1-based from R, converted to 0-based)
    nn : int
        Second parameter index (1-based from R, converted to 0-based)
        
    Returns
    -------
    float
        Second derivative component
    """
    # Convert from R's 1-based to Python's 0-based indexing
    mm = mm - 1
    nn = nn - 1
    
    d1 = fd1 * v1
    d2 = fd2 * v2
    net3 = np.zeros((3, 2))
    net4 = np.zeros((4, 2))
    lmn = np.zeros(4)
    dd = np.array([d1, d2])
    vv = np.array([v1, v2])
    vvd = np.zeros(2)
    nx = len(x)
    
    # different
    if mm != nn:
        net4[:, mm] = [-1, -1, 1, 1]
        net4[:, nn] = [-1, 1, -1, 1]
        for i in range(4):
            for j in range(2):
                vvd[j] = vv[j] + net4[i, j] * dd[j]
            lmn[i] = np.sum(stats.gamma.logpdf(x, a=vvd[0], scale=vvd[1])) / nx
        dld = (lmn[0] - lmn[1] - lmn[2] + lmn[3]) / (4 * dd[mm] * dd[nn])
    # same
    else:
        net3[:, mm] = [-1, 0, 1]
        for i in range(3):
            for j in range(2):
                vvd[j] = vv[j] + net3[i, j] * dd[j]
            lmn[i] = np.sum(stats.gamma.logpdf(x, a=vvd[0], scale=vvd[1])) / nx
        dld = (lmn[0] - 2 * lmn[1] + lmn[2]) / (dd[mm] * dd[mm])
    
    return dld

def gamma_ldd(x, v1, fd1, v2, fd2):
    """
    Second derivative matrix of the normalized log-likelihood
    
    Parameters
    ----------
    x : array-like
        Data values
    v1 : float
        Shape parameter
    fd1 : float
        Step size for v1 finite differences
    v2 : float
        Scale parameter
    fd2 : float
        Step size for v2 finite differences
        
    Returns
    -------
    ndarray
        2x2 second derivative matrix
    """
    ldd = np.zeros((2, 2))
    for i in range(1, 3):  # R's 1:2
        for j in range(i, 3):  # R's i:2
            ldd[i-1, j-1] = gamma_lmn(x, v1, fd1, v2, fd2, i, j)
    
    for i in range(2, 0, -1):  # R's 2:1
        for j in range(1, i):  # R's 1:(i-1)
            ldd[i-1, j-1] = ldd[j-1, i-1]
    
    return ldd

def gamma_gmn(alpha, v1, fd1, v2, fd2, mm, nn):
    """
    One component of the second derivative of the expected log-likelihood
    
    Parameters
    ----------
    alpha : array-like
        Alpha values
    v1 : float
        Shape parameter
    fd1 : float
        Step size for v1 finite differences
    v2 : float
        Scale parameter
    fd2 : float
        Step size for v2 finite differences
    mm : int
        First parameter index (1-based from R, converted to 0-based)
    nn : int
        Second parameter index (1-based from R, converted to 0-based)
        
    Returns
    -------
    ndarray
        Derivative component values
    """
    # Convert from R's 1-based to Python's 0-based indexing
    mm = mm - 1
    nn = nn - 1
    
    nx = len(alpha)
    d1 = fd1 * v1
    d2 = fd2 * v2
    x = stats.gamma.ppf(1 - alpha, a=v1, scale=v2)
    net3 = np.zeros((3, 2))
    net4 = np.zeros((4, 2))
    lmn = np.zeros((nx, 4))
    dd = np.array([d1, d2])
    vv = np.array([v1, v2])
    vvd = np.zeros(2)
    nx = len(x)
    
    # different
    if mm != nn:
        net4[:, mm] = [-1, -1, 1, 1]
        net4[:, nn] = [-1, 1, -1, 1]
        for i in range(4):
            for j in range(2):
                vvd[j] = vv[j] + net4[i, j] * dd[j]
            lmn[:, i] = stats.gamma.logpdf(x, a=vvd[0], scale=vvd[1])
        dld = (lmn[:, 0] - lmn[:, 1] - lmn[:, 2] + lmn[:, 3]) / (4 * dd[mm] * dd[nn])
    # same
    else:
        net3[:, mm] = [-1, 0, 1]
        for i in range(3):
            for j in range(2):
                vvd[j] = vv[j] + net3[i, j] * dd[j]
            lmn[:, i] = stats.gamma.logpdf(x, a=vvd[0], scale=vvd[1])
        dld = (lmn[:, 0] - 2 * lmn[:, 1] + lmn[:, 2]) / (dd[mm] * dd[mm])
    
    return dld

def gamma_gg(v1, fd1, v2, fd2):
    """
    Second derivative matrix of the expected log-likelihood
    
    Parameters
    ----------
    v1 : float
        Shape parameter
    fd1 : float
        Step size for v1 finite differences
    v2 : float
        Scale parameter
    fd2 : float
        Step size for v2 finite differences
        
    Returns
    -------
    ndarray
        2x2 expected information matrix
    """
    expinfmat = np.zeros((2, 2))
    for i in range(1, 3):  # R's 1:2
        for j in range(i, 3):  # R's i:2
            result, _ = integrate.quad(gamma_gmn, 0, 1, args=(v1, fd1, v2, fd2, i, j))
            expinfmat[i-1, j-1] = -result
    
    for i in range(2, 2+1):  # R's 2:2
        for j in range(1, i):  # R's 1:(i-1)
            expinfmat[i-1, j-1] = expinfmat[j-1, i-1]
    
    return expinfmat

def gamma_lmnp(x, v1, fd1, v2, fd2, mm, nn, rr):
    """
    One component of the third derivative of the normalized log-likelihood
    
    Parameters
    ----------
    x : array-like
        Data values
    v1 : float
        Shape parameter
    fd1 : float
        Step size for v1 finite differences
    v2 : float
        Scale parameter
    fd2 : float
        Step size for v2 finite differences
    mm : int
        First parameter index (1-based from R, converted to 0-based)
    nn : int
        Second parameter index (1-based from R, converted to 0-based)
    rr : int
        Third parameter index (1-based from R, converted to 0-based)
        
    Returns
    -------
    float
        Third derivative component
    """
    # Convert from R's 1-based to Python's 0-based indexing
    mm = mm - 1
    nn = nn - 1
    rr = rr - 1
    
    d1 = fd1 * v1
    d2 = fd2 * v2
    net4 = np.zeros((4, 2))
    net6 = np.zeros((6, 2))
    net8 = np.zeros((8, 2))
    lmn = np.zeros(8)
    dd = np.array([d1, d2])
    vv = np.array([v1, v2])
    vvd = np.zeros(2)
    nx = len(x)
    
    # all diff
    if (mm != nn) and (nn != rr) and (rr != mm):
        net8[:, mm] = [-1, 1, -1, 1, -1, 1, -1, 1]
        net8[:, nn] = [-1, -1, 1, 1, -1, -1, 1, 1]
        net8[:, rr] = [-1, -1, -1, -1, 1, 1, 1, 1]
        for i in range(8):
            for j in range(2):  # Only 2 parameters, not 3
                vvd[j] = vv[j] + net8[i, j] * dd[j]
            lmn[i] = np.sum(stats.gamma.logpdf(x, a=vvd[0], scale=vvd[1])) / nx
        dld1 = (lmn[1] - lmn[0]) / (2 * dd[mm])
        dld2 = (lmn[3] - lmn[2]) / (2 * dd[mm])
        dld21 = (dld2 - dld1) / (2 * dd[nn])
        dld3 = (lmn[5] - lmn[4]) / (2 * dd[mm])
        dld4 = (lmn[7] - lmn[6]) / (2 * dd[mm])
        dld43 = (dld4 - dld3) / (2 * dd[nn])
        dld = (dld43 - dld21) / (2 * dd[rr])
    # all 3 the same
    elif (mm == nn) and (nn == rr):
        net4[:, mm] = [-2, -1, 1, 2]
        for i in range(4):
            for j in range(2):
                vvd[j] = vv[j] + net4[i, j] * dd[j]
            lmn[i] = np.sum(stats.gamma.logpdf(x, a=vvd[0], scale=vvd[1])) / nx
        dld = (-lmn[0] + 2 * lmn[1] - 2 * lmn[2] + lmn[3]) / (2 * dd[mm] * dd[mm] * dd[mm])
    else:
        # 2 the same
        # mm is the repeated one, nn is the other one
        if mm == nn:
            m2 = mm
            n2 = rr
        elif mm == rr:
            m2 = mm
            n2 = nn
        elif nn == rr:
            m2 = nn
            n2 = mm
        net6[:, m2] = [-1, 0, 1, -1, 0, 1]
        net6[:, n2] = [-1, -1, -1, 1, 1, 1]
        for i in range(6):
            for j in range(2):
                vvd[j] = vv[j] + net6[i, j] * dd[j]
            lmn[i] = np.sum(stats.gamma.logpdf(x, a=vvd[0], scale=vvd[1])) / nx
        dld1 = (lmn[2] - 2 * lmn[1] + lmn[0]) / (dd[m2] * dd[m2])
        dld2 = (lmn[5] - 2 * lmn[4] + lmn[3]) / (dd[m2] * dd[m2])
        dld = (dld2 - dld1) / (2 * dd[n2])
    
    return dld

def gamma_lddd(x, v1, fd1, v2, fd2):
    """
    Third derivative tensor of the normalized log-likelihood
    
    Parameters
    ----------
    x : array-like
        Data values
    v1 : float
        Shape parameter
    fd1 : float
        Step size for v1 finite differences
    v2 : float
        Scale parameter
    fd2 : float
        Step size for v2 finite differences
        
    Returns
    -------
    ndarray
        2x2x2 third derivative tensor
    """
    # calculate the unique values
    lddd = np.zeros((2, 2, 2))
    for i in range(1, 3):  # R's 1:2
        for j in range(i, 3):  # R's i:2
            for k in range(j, 3):  # R's j:2
                lddd[i-1, j-1, k-1] = gamma_lmnp(x, v1, fd1, v2, fd2, i, j, k)
    
    # steves dumb algorithm for filling in the non-unique values
    for i in range(2):
        for j in range(2):
            for k in range(2):
                a = [i, j, k]
                b = sorted(a)
                lddd[a[0], a[1], a[2]] = lddd[b[0], b[1], b[2]]
    
    return lddd

def gamma_f1f(y, v1, fd1, v2, fd2):
    """
    DMGS equation 3.3, f1 term
    
    Parameters
    ----------
    y : array-like
        Data values
    v1 : float
        Shape parameter
    fd1 : float
        Step size for v1 finite differences
    v2 : float
        Scale parameter
    fd2 : float
        Step size for v2 finite differences
        
    Returns
    -------
    ndarray
        2 x len(y) matrix of first derivatives
    """
    d1 = fd1 * v1
    d2 = fd2 * v2
    # v1 stuff
    v1m1 = v1 - 1 * d1
    v100 = v1 + 0 * d1
    v1p1 = v1 + 1 * d1
    # v2 stuff
    v2m1 = v2 - 1 * d2
    v200 = v2 + 0 * d2
    v2p1 = v2 + 1 * d2
    # v1 derivatives
    F1m1 = stats.gamma.pdf(y, a=v1m1, scale=v200)
    F1p1 = stats.gamma.pdf(y, a=v1p1, scale=v200)
    # v2 derivatives
    F2m1 = stats.gamma.pdf(y, a=v100, scale=v2m1)
    F2p1 = stats.gamma.pdf(y, a=v100, scale=v2p1)
    f1 = np.zeros((2, len(y)))
    f1[0, :] = (F1p1 - F1m1) / (2 * d1)
    f1[1, :] = (F2p1 - F2m1) / (2 * d2)
    return f1

def gamma_p1f(y, v1, fd1, v2, fd2):
    """
    DMGS equation 3.3, p1 term
    
    Parameters
    ----------
    y : array-like
        Data values
    v1 : float
        Shape parameter
    fd1 : float
        Step size for v1 finite differences
    v2 : float
        Scale parameter
    fd2 : float
        Step size for v2 finite differences
        
    Returns
    -------
    ndarray
        2 x len(y) matrix of first derivatives of CDF
    """
    d1 = fd1 * v1
    d2 = fd2 * v2
    # v1 stuff
    v1m1 = v1 - 1 * d1
    v100 = v1 + 0 * d1
    v1p1 = v1 + 1 * d1
    # v2 stuff
    v2m1 = v2 - 1 * d2
    v200 = v2 + 0 * d2
    v2p1 = v2 + 1 * d2
    # v1 derivatives
    F1m1 = stats.gamma.cdf(y, a=v1m1, scale=v200)
    F1p1 = stats.gamma.cdf(y, a=v1p1, scale=v200)
    # v2 derivatives
    F2m1 = stats.gamma.cdf(y, a=v100, scale=v2m1)
    F2p1 = stats.gamma.cdf(y, a=v100, scale=v2p1)
    p1 = np.zeros((2, len(y)))
    p1[0, :] = (F1p1 - F1m1) / (2 * d1)
    p1[1, :] = (F2p1 - F2m1) / (2 * d2)
    return p1

def gamma_mu1f(alpha, v1, fd1, v2, fd2):
    """
    DMGS equation 3.3, mu1 term
    
    Parameters
    ----------
    alpha : array-like
        Alpha values
    v1 : float
        Shape parameter
    fd1 : float
        Step size for v1 finite differences
    v2 : float
        Scale parameter
    fd2 : float
        Step size for v2 finite differences
        
    Returns
    -------
    ndarray
        2 x len(alpha) matrix of mu1 values
    """
    q00 = stats.gamma.ppf(1 - alpha, a=v1, scale=v2)
    d1 = fd1 * v1
    d2 = fd2 * v2
    # v1 stuff
    v1m1 = v1 - 1 * d1
    v100 = v1 + 0 * d1
    v1p1 = v1 + 1 * d1
    # v2 stuff
    v2m1 = v2 - 1 * d2
    v200 = v2 + 0 * d2
    v2p1 = v2 + 1 * d2
    # v1 derivatives
    F1m1 = stats.gamma.cdf(q00, a=v1m1, scale=v200)
    F1p1 = stats.gamma.cdf(q00, a=v1p1, scale=v200)
    # v2 derivatives
    F2m1 = stats.gamma.cdf(q00, a=v100, scale=v2m1)
    F2p1 = stats.gamma.cdf(q00, a=v100, scale=v2p1)
    mu1 = np.zeros((2, len(alpha)))
    mu1[0, :] = -(F1p1 - F1m1) / (2 * d1)
    mu1[1, :] = -(F2p1 - F2m1) / (2 * d2)
    return mu1

def gamma_f2f(y, v1, fd1, v2, fd2):
    """
    DMGS equation 3.3, f2 term
    
    Parameters
    ----------
    y : array-like
        Data values
    v1 : float
        Shape parameter
    fd1 : float
        Step size for v1 finite differences
    v2 : float
        Scale parameter
    fd2 : float
        Step size for v2 finite differences
        
    Returns
    -------
    ndarray
        2 x 2 x len(y) array of second derivatives
    """
    d1 = fd1 * v1
    d2 = fd2 * v2
    # v1 stuff
    v1m2 = v1 - 2 * d1
    v1m1 = v1 - 1 * d1
    v100 = v1 + 0 * d1
    v1p1 = v1 + 1 * d1
    v1p2 = v1 + 2 * d1
    # v2 stuff
    v2m2 = v2 - 2 * d2
    v2m1 = v2 - 1 * d2
    v200 = v2 + 0 * d2
    v2p1 = v2 + 1 * d2
    v2p2 = v2 + 2 * d2
    F1m2 = stats.gamma.pdf(y, a=v1m2, scale=v200)
    F1m1 = stats.gamma.pdf(y, a=v1m1, scale=v200)
    F100 = stats.gamma.pdf(y, a=v100, scale=v200)
    F1p1 = stats.gamma.pdf(y, a=v1p1, scale=v200)
    F1p2 = stats.gamma.pdf(y, a=v1p2, scale=v200)
    # v2 derivative
    F2m2 = stats.gamma.pdf(y, a=v100, scale=v2m2)
    F2m1 = stats.gamma.pdf(y, a=v100, scale=v2m1)
    F200 = stats.gamma.pdf(y, a=v100, scale=v200)
    F2p1 = stats.gamma.pdf(y, a=v100, scale=v2p1)
    F2p2 = stats.gamma.pdf(y, a=v100, scale=v2p2)
    # cross derivative
    Fcm1m1 = stats.gamma.pdf(y, a=v1m1, scale=v2m1)
    Fcm1p1 = stats.gamma.pdf(y, a=v1m1, scale=v2p1)
    Fcp1m1 = stats.gamma.pdf(y, a=v1p1, scale=v2m1)
    Fcp1p1 = stats.gamma.pdf(y, a=v1p1, scale=v2p1)
    f2 = np.zeros((2, 2, len(y)))
    f2[0, 0, :] = (F1p1 - 2 * F100 + F1m1) / (d1 * d1)
    f2[1, 1, :] = (F2p1 - 2 * F200 + F2m1) / (d2 * d2)
    f2[0, 1, :] = (Fcp1p1 - Fcm1p1 - Fcp1m1 + Fcm1m1) / (4 * d1 * d2)
    # copy
    f2[1, 0, :] = f2[0, 1, :]
    return f2

def gamma_p2f(y, v1, fd1, v2, fd2):
    """
    DMGS equation 3.3, p2 term
    
    Parameters
    ----------
    y : array-like
        Data values
    v1 : float
        Shape parameter
    fd1 : float
        Step size for v1 finite differences
    v2 : float
        Scale parameter
    fd2 : float
        Step size for v2 finite differences
        
    Returns
    -------
    ndarray
        2 x 2 x len(y) array of second derivatives of CDF
    """
    d1 = fd1 * v1
    d2 = fd2 * v2
    # v1 stuff
    v1m2 = v1 - 2 * d1
    v1m1 = v1 - 1 * d1
    v100 = v1 + 0 * d1
    v1p1 = v1 + 1 * d1
    v1p2 = v1 + 2 * d1
    # v2 stuff
    v2m2 = v2 - 2 * d2
    v2m1 = v2 - 1 * d2
    v200 = v2 + 0 * d2
    v2p1 = v2 + 1 * d2
    v2p2 = v2 + 2 * d2
    F1m2 = stats.gamma.cdf(y, a=v1m2, scale=v200)
    F1m1 = stats.gamma.cdf(y, a=v1m1, scale=v200)
    F100 = stats.gamma.cdf(y, a=v100, scale=v200)
    F1p1 = stats.gamma.cdf(y, a=v1p1, scale=v200)
    F1p2 = stats.gamma.cdf(y, a=v1p2, scale=v200)
    # v2 derivative
    F2m2 = stats.gamma.cdf(y, a=v100, scale=v2m2)
    F2m1 = stats.gamma.cdf(y, a=v100, scale=v2m1)
    F200 = stats.gamma.cdf(y, a=v100, scale=v200)
    F2p1 = stats.gamma.cdf(y, a=v100, scale=v2p1)
    F2p2 = stats.gamma.cdf(y, a=v100, scale=v2p2)
    # cross derivative
    Fcm1m1 = stats.gamma.cdf(y, a=v1m1, scale=v2m1)
    Fcm1p1 = stats.gamma.cdf(y, a=v1m1, scale=v2p1)
    Fcp1m1 = stats.gamma.cdf(y, a=v1p1, scale=v2m1)
    Fcp1p1 = stats.gamma.cdf(y, a=v1p1, scale=v2p1)
    p2 = np.zeros((2, 2, len(y)))
    p2[0, 0, :] = (F1p1 - 2 * F100 + F1m1) / (d1 * d1)
    p2[1, 1, :] = (F2p1 - 2 * F200 + F2m1) / (d2 * d2)
    p2[0, 1, :] = (Fcp1p1 - Fcm1p1 - Fcp1m1 + Fcm1m1) / (4 * d1 * d2)
    # copy
    p2[1, 0, :] = p2[0, 1, :]
    return p2

def gamma_mu2f(alpha, v1, fd1, v2, fd2):
    """
    DMGS equation 3.3, mu2 term
    
    Parameters
    ----------
    alpha : array-like
        Alpha values
    v1 : float
        Shape parameter
    fd1 : float
        Step size for v1 finite differences
    v2 : float
        Scale parameter
    fd2 : float
        Step size for v2 finite differences
        
    Returns
    -------
    ndarray
        2 x 2 x len(alpha) array of mu2 values
    """
    q00 = stats.gamma.ppf(1 - alpha, a=v1, scale=v2)
    d1 = fd1 * v1
    d2 = fd2 * v2
    # v1 stuff
    v1m2 = v1 - 2 * d1
    v1m1 = v1 - 1 * d1
    v100 = v1 + 0 * d1
    v1p1 = v1 + 1 * d1
    v1p2 = v1 + 2 * d1
    # v2 stuff
    v2m2 = v2 - 2 * d2
    v2m1 = v2 - 1 * d2
    v200 = v2 + 0 * d2
    v2p1 = v2 + 1 * d2
    v2p2 = v2 + 2 * d2
    mu2 = np.zeros((2, 2, len(alpha)))
    F1m2 = stats.gamma.cdf(q00, a=v1m2, scale=v200)
    F1m1 = stats.gamma.cdf(q00, a=v1m1, scale=v200)
    F100 = stats.gamma.cdf(q00, a=v100, scale=v200)
    F1p1 = stats.gamma.cdf(q00, a=v1p1, scale=v200)
    F1p2 = stats.gamma.cdf(q00, a=v1p2, scale=v200)
    # v2 derivative
    F2m2 = stats.gamma.cdf(q00, a=v100, scale=v2m2)
    F2m1 = stats.gamma.cdf(q00, a=v100, scale=v2m1)
    F200 = stats.gamma.cdf(q00, a=v100, scale=v200)
    F2p1 = stats.gamma.cdf(q00, a=v100, scale=v2p1)
    F2p2 = stats.gamma.cdf(q00, a=v100, scale=v2p2)
    # cross derivative
    Fcm1m1 = stats.gamma.cdf(q00, a=v1m1, scale=v2m1)
    Fcm1p1 = stats.gamma.cdf(q00, a=v1m1, scale=v2p1)
    Fcp1m1 = stats.gamma.cdf(q00, a=v1p1, scale=v2m1)
    Fcp1p1 = stats.gamma.cdf(q00, a=v1p1, scale=v2p1)
    mu2[0, 0, :] = -(F1p1 - 2 * F100 + F1m1) / (d1 * d1)
    mu2[1, 1, :] = -(F2p1 - 2 * F200 + F2m1) / (d2 * d2)
    mu2[0, 1, :] = -(Fcp1p1 - Fcm1p1 - Fcp1m1 + Fcm1m1) / (4 * d1 * d2)
    # copy
    mu2[1, 0, :] = mu2[0, 1, :]
    return mu2

def gamma_means(means, ml_params, lddi, lddd, lambdad_cp, nx, dim=2):
    """
    MLE and RHP predictive means
    
    Parameters
    ----------
    means : bool
        Whether to calculate means
    ml_params : array-like
        Maximum likelihood parameter estimates [shape, scale]
    lddi : ndarray
        Inverse of second derivative matrix
    lddd : ndarray
        Third derivative tensor
    lambdad_cp : array-like
        Lambda parameter vector
    nx : int
        Sample size
    dim : int
        Number of parameters (default 2)
        
    Returns
    -------
    dict
        Dictionary with ml_mean and cp_mean
    """
    # v1 is shape
    # v2 is scale
    if means:
        v1 = ml_params[0]
        v2 = ml_params[1]
        
        # ml mean
        ml_mean = v1 * v2
        
        # cp mean
        meand1 = np.zeros((2, 1))
        meand1[0, 0] = v2
        meand1[1, 0] = v1
        
        meand2 = np.zeros((2, 2, 1))
        meand2[0, 0, 0] = 0
        meand2[0, 1, 0] = 1
        meand2[1, 0, 0] = 1
        meand2[1, 1, 0] = 0
        
        dmean = cp_dmgs.dmgs(lddi, lddd, meand1, lambdad_cp, meand2, dim=2)
        cp_mean = ml_mean + dmean / nx
    else:
        ml_mean = "means not selected"
        cp_mean = "means not selected"
    
    return {'ml_mean': ml_mean, 'cp_mean': cp_mean}

def gamma_logscores(logscores, x, fd1=0.01, fd2=0.01, aderivs=True):
    """
    Log scores for MLE and RHP predictions calculated using leave-one-out
    
    Parameters
    ----------
    logscores : bool
        Whether to calculate log scores
    x : array-like
        Data values
    fd1 : float
        Step size for v1 finite differences (default 0.01)
    fd2 : float
        Step size for v2 finite differences (default 0.01)
    aderivs : bool
        Whether to use analytical derivatives (default True)
        
    Returns
    -------
    dict
        Dictionary with ml_oos_logscore and cp_oos_logscore
    """
    if logscores:
        nx = len(x)
        ml_oos_logscore = 0
        cp_oos_logscore = 0
        for i in range(nx):
            x1 = np.delete(x, i)
            
            dd = dgammasub(x1, x[i], fd1, fd2, aderivs)
            
            ml_params = dd['ml_params']
            
            ml_pdf = dd['ml_pdf']
            ml_oos_logscore = ml_oos_logscore + np.log(ml_pdf)
            
            cp_pdf = dd['cp_pdf']
            cp_oos_logscore = cp_oos_logscore + np.log(max(cp_pdf, sys.float_info.epsilon))
    else:
        ml_oos_logscore = "logscores not selected"
        cp_oos_logscore = "logscores not selected"
    
    return {'ml_oos_logscore': ml_oos_logscore, 'cp_oos_logscore': cp_oos_logscore}


def dgammasub(x, y, fd1=0.01, fd2=0.01, aderivs=True):
    """
    Densities from MLE and RHP
    
    Parameters
    ----------
    x : array-like
        Training data values
    y : float
        Test point
    fd1 : float
        Step size for v1 finite differences (default 0.01)
    fd2 : float
        Step size for v2 finite differences (default 0.01)
    aderivs : bool
        Whether to use analytical derivatives (default True)
        
    Returns
    -------
    dict
        Dictionary with ml_params, ml_pdf, cp_pdf, ml_cdf, cp_cdf
    """
    nx = len(x)
    
    opt = optimize.minimize(lambda params: -gamma_loglik(params, x), 
                          x0=[1, 1], method='BFGS')
    v1hat = opt.x[0]
    v2hat = opt.x[1]
    ml_params = np.array([v1hat, v2hat])
    
    # mle
    ml_pdf = stats.gamma.pdf(y, a=v1hat, scale=v2hat)
    ml_cdf = stats.gamma.cdf(y, a=v1hat, scale=v2hat)
    
    # cp
    if aderivs:
        ldd = gamma_derivs.gamma_ldda(x, v1hat, v2hat)
    else:
        ldd = gamma_ldd(x, v1hat, fd1, v2hat, fd2)
    lddi = np.linalg.inv(ldd)
    
    if aderivs:
        lddd = gamma_derivs.gamma_lddda(x, v1hat, v2hat)
    else:
        lddd = gamma_lddd(x, v1hat, fd1, v2hat, fd2)
    
    if aderivs:
        f1 = gamma_derivs.gamma_f1fa(y, v1hat, v2hat)
    else:
        f1 = gamma_f1f(y, v1hat, fd1, v2hat, fd2)
    
    if aderivs:
        f2 = gamma_derivs.gamma_f2fa(y, v1hat, v2hat)
    else:
        f2 = gamma_f2f(y, v1hat, fd1, v2hat, fd2)
    
    p1 = gamma_p1f(y, v1hat, fd1, v2hat, fd2)
    p2 = gamma_p2f(y, v1hat, fd1, v2hat, fd2)
    lambdad_cp = np.array([-1/v1hat, -1/v2hat])
    df1 = cp_dmgs.dmgs(lddi, lddd, f1, lambdad_cp, f2, dim=2)
    dp1 = cp_dmgs.dmgs(lddi, lddd, p1, lambdad_cp, p2, dim=2)
    cp_pdf = np.maximum(ml_pdf + df1/nx, 0)
    cp_cdf = np.minimum(np.maximum(ml_cdf + dp1/nx, 0), 1)
    
    # return
    return {
        'ml_params': ml_params,
        'ml_pdf': ml_pdf,
        'cp_pdf': cp_pdf,
        'ml_cdf': ml_cdf,
        'cp_cdf': cp_cdf
    }


import scipy.stats as sp
import sympy
import autodiff

x, v1, v2 = sympy.symbols('x v1 v2')

f = (1/sympy.sqrt(2*sympy.pi))*(1/v2)*(1/x)*sympy.exp(-(sympy.log(x)-v1)**2/(2*v2*v2))
logf = -0.5*sympy.log(2*sympy.pi)-sympy.log(v2)-sympy.log(x)-(sympy.log(x)-v1)**2/(2*v2*v2)

                                 
with open('output_lnormal.py', 'w') as file:
    file.write('import numpy as np\n')
    file.write(autodiff.diff_to_code(f, [v1, v2], 'lnorm_fd', 1))
    file.write(autodiff.diff_to_code(f, [v1, v2], 'lnorm_fdd', 2))
    file.write(autodiff.diff_to_code(logf, [v1, v2], 'lnorm_logfdd', 2))
    file.write(autodiff.diff_to_code(logf, [v1, v2], 'lnorm_logfddd', 3))

file.close()

# lnorm_pd and lnorm_pdd done manually


'''
fitdistcp package. Visit fitdistcp.info for more information.
Provides modules 
    - genextreme: generalised extreme value distribution
    - genextreme_p1: generalised extreme value distribution, with one predictor
    - genpareto: generalised pareto distribution
'''

__all__ = [
    'genextreme', 
    'genextreme_p1', 
    'genpareto',
    'expon',
	'gamma',
	'gumbel',
	'lnorm',
	'norm',
	'weibull',
    'genextreme_p12',
    'reltest'
]


import scipy
import sympy
import numpy as np
import autodiff

x, v1= sympy.symbols('x v1')

f = v1*sympy.exp(-v1*x)
p = 1-sympy.exp(-v1*x)
logf = sympy.log(v1)-v1*x
                                 
with open('output_exp.py', 'w') as file:
    file.write('import numpy as np\n')
    file.write(autodiff.diff_to_code(f, [v1], 'exp_fd', 1))
    file.write(autodiff.diff_to_code(f, [v1], 'exp_fdd', 2))
    file.write(autodiff.diff_to_code(p, [v1], 'exp_pd', 1))
    file.write(autodiff.diff_to_code(p, [v1], 'exp_pdd', 2))
    file.write(autodiff.diff_to_code(logf, [v1], 'exp_logfdd', 2))
    file.write(autodiff.diff_to_code(logf, [v1], 'exp_logfddd', 3))

file.close()


import numpy as np


def bayesian_dq_4terms_v1(lddi, lddd, mu1, pidopi1, pidopi2, mu2, dim):
    """
    Evaluate DMGS equation 3.3
    
    Parameters:
    lddi: inverse of second derivative of observed log-likelihood
    lddd: third derivative of observed log-likelihood  
    mu1: DMGS mu1 vector
    pidopi1: first part of the prior term
    pidopi2: second part of the prior term
    mu2: DMGS mu2 matrix
    dim: number of parameters
    
    Returns:
    Vector
    """
    nalpha = mu1.shape[1]  # length(mu1[1,])
    beps1 = np.zeros(nalpha)  # rep(0, nalpha)
    beps2 = np.zeros(nalpha)
    beps3 = np.zeros(nalpha)
    beps4 = np.zeros(nalpha)
    
    # beps1
    for t in range(dim):  # 1:dim becomes 0:dim
        for j in range(dim):
            for r in range(dim):
                for s in range(dim):
                    beps1 = beps1 + 0.5 * lddi[s, t] * lddi[j, r] * lddd[j, r, s] * mu1[t, :]
    
    # beps2
    for t in range(dim):
        for s in range(dim):
            beps2 = beps2 - lddi[s, t] * pidopi1[s] * mu1[t, :]  # this has to be negative because ldd is the opposite sign to c
    
    # beps3
    for t in range(dim):
        for s in range(dim):
            beps3 = beps3 - lddi[s, t] * pidopi2[s] * mu1[t, :]  # this has to be negative because ldd is the opposite sign to c
    
    # beps4
    for j in range(dim):
        for r in range(dim):
            beps4 = beps4 - 0.5 * lddi[j, r] * mu2[j, r, :]  # this has to be negative because ldd is the opposite sign to c
    
    beps = beps1 + beps2 + beps3 + beps4
    return beps


def dmgs(lddi, lddd, mu1, pidopi, mu2, dim):
    # the inputs must have the correct number of dimensions, even if the last one is empty:
    if isinstance(pidopi, float):
        pidopi = np.asarray([pidopi])
    if len(mu1.shape) == 1:     
        mu1 = np.expand_dims(mu1, axis=-1)
    if len(mu2.shape) == 2:
        mu2 = np.expand_dims(mu2, axis=-1)

    pidopi1 = 2*pidopi
    pidopi2 = -1*pidopi
    
    nalpha = mu1.shape[1]
    beps1 = np.zeros(nalpha)
    beps2 = np.zeros(nalpha)
    beps3 = np.zeros(nalpha)
    beps4 = np.zeros(nalpha)
    # beps1
    for t in range(dim):
        for j in range(dim):
            for r in range(dim):
                for s in range(dim):
                    beps1 = beps1+0.5 * lddi[s, t] * lddi[j, r] * lddd[j, r, s] * mu1[t, :]
    
    # beps2
    for t in range(dim):
        for s in range(dim):
            beps2 = beps2- lddi[s, t] * pidopi1[s] * mu1[t, :]  # this has to be negative because ldd is the opposite sign to c
    
    # beps3
    for t in range(dim):
        for s in range(dim):
            beps3 = beps3- lddi[s, t] * pidopi2[s] * mu1[t, :]  # this has to be negative because ldd is the opposite sign to c
    
    # beps4
    for j in range(dim):
        for r in range(dim):
            beps4 = beps4 - 0.5 * lddi[j, r] * mu2[j, r, :]  # this has to be negative because ldd is the opposite sign to c
    
    beps = beps1 + beps2 + beps3 + beps4
    return beps



import numpy as np
import scipy.stats
import sys

from . import utils as cp_utils
from . import lnorm_derivs


def norm_ml_params(x):
    """Maximum likelihood parameters for normal distribution"""
    n = len(x)
    mean_est = np.mean(x)
    var_est = np.sum((x - mean_est)**2) / n
    sd_est = np.sqrt(var_est)
    return np.array([mean_est, sd_est])


def lnorm_waic(waicscores, x, v1hat, v2hat):
    """
    Waic for RUST
    
    Parameters
    ----------
    waicscores : bool
        Whether to calculate WAIC scores
    x : array-like
        Data values
    v1hat : float
        Estimated meanlog parameter
    v2hat : float
        Estimated sdlog parameter
    
    Returns
    -------
    dict
        Dictionary containing waic1 and waic2 scores
    """
    if waicscores:
        f1f = lnorm_derivs.lnorm_f1fa(x, v1hat, v2hat)
        f2f = lnorm_derivs.lnorm_f2fa(x, v1hat, v2hat)
        
        ldd = lnorm_derivs.lnorm_ldda(x, v1hat, v2hat)
        lddi = np.linalg.inv(ldd)
        lddd = lnorm_derivs.lnorm_lddda(x, v1hat, v2hat)
        fhatx = scipy.stats.lognorm.pdf(x, s=v2hat, scale=np.exp(v1hat))
        lambdad_rhp = np.array([0, -1/v2hat])
        waic = cp_utils.make_waic(x, fhatx, lddi, lddd, f1f, lambdad_rhp, f2f, dim=2)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
    else:
        waic1 = "waicscores not selected"
        waic2 = "waicscores not selected"
    
    return {'waic1': waic1, 'waic2': waic2}


def lnorm_logf(params, x):
    """
    Logf for RUST
    
    Parameters
    ----------
    params : array-like
        Parameters [meanlog, sdlog]
    x : array-like
        Data values
    
    Returns
    -------
    float
        Log-likelihood value
    """
    m = params[0]
    s = np.maximum(params[1], sys.float_info.epsilon)
    logf = np.sum(scipy.stats.lognorm.logpdf(x, s=s, scale=np.exp(m))) - np.log(s)
    return logf


def lnorm_logscores(logscores, x):
    """
    Log scores for MLE and RHP predictions calculated using leave-one-out
    
    Parameters
    ----------
    logscores : bool
        Whether to calculate log scores
    x : array-like
        Data values
    
    Returns
    -------
    dict
        Dictionary containing ml_oos_logscore and rh_oos_logscore
    """
    if logscores:
        y = np.log(x)
        nx = len(x)
        ml_oos_logscore = 0
        rh_oos_logscore = 0
        
        for i in range(nx):
            # Create leave-one-out dataset
            x1 = np.concatenate([x[:i], x[i+1:]])
            
            dd = dlnormsub(x1, x[i])
            
            ml_params = dd['ml_params']
            
            ml_pdf = dd['ml_pdf']
            ml_oos_logscore = ml_oos_logscore + np.log(ml_pdf)
            
            rh_pdf = dd['rh_pdf']
            rh_oos_logscore = rh_oos_logscore + np.log(rh_pdf)
    else:
        ml_oos_logscore = "logscores not selected"
        rh_oos_logscore = "logscores not selected"
    
    return {'ml_oos_logscore': ml_oos_logscore, 'rh_oos_logscore': rh_oos_logscore}


def dlnormsub(x, y):
    """
    Densities from MLE and RHP
    
    Parameters
    ----------
    x : array-like
        Training data values
    y : float
        Test data value
    
    Returns
    -------
    dict
        Dictionary containing ml_params, ml_pdf, rh_pdf, ml_cdf, rh_cdf
    """
    # great potential for confusion here
    # x and y are both lognormal
    # logx and logy are both normal
    
    nx = len(x)
    logx = np.log(x)
    logy = np.log(y)
    
    # ml
    ml_params = norm_ml_params(logx)  # this really should be norm not lnorm
    ml_pdf = scipy.stats.lognorm.pdf(y, s=ml_params[1], scale=np.exp(ml_params[0]))
    ml_cdf = scipy.stats.lognorm.cdf(y, s=ml_params[1], scale=np.exp(ml_params[0]))
    
    # rhp pdf
    # -first, convert sigma from maxlik to unbiased
    sgu = ml_params[1] * np.sqrt(nx/(nx-1))
    # then, convert sigma to predictive sigma
    sg1 = sgu * np.sqrt((nx+1)/nx)
    
    logyd = (logy - ml_params[0]) / sg1
    rh_pdf = (scipy.stats.t.pdf(logyd, df=nx-1)) / (y * sg1)  # note the extra exp term here, to convert to loglnormal predictive density
    rh_pdf = np.maximum(rh_pdf, 0)
    
    # rhp cdf
    rh_cdf = scipy.stats.t.cdf(logyd, df=nx-1)
    rh_cdf = np.minimum(np.maximum(rh_cdf, 0), 1)
    
    # return
    return {
        'ml_params': ml_params,
        'ml_pdf': ml_pdf,
        'rh_pdf': rh_pdf,
        'ml_cdf': ml_cdf,
        'rh_cdf': rh_cdf
    }




import numpy as np
import sys
import textwrap

from . import evaluate_dmgs_equation as cp_dmgs

 
# Deal with situations in which the user wants d or p outside the GEV range
def fixgevrange(y, v1, v2, v3):
    # I've added minxi now, so that even in the case where I adjust xi slightly
    # in the 'movexiawayfromzero' routine, to avoid being too near zero when
    # calculating analytical gradients, we still adjust the y and we don't
    # get NA errors from the analytical gradient routines, from logp1
    minxi = 10**-7
    y = np.array(y, copy=True)
    
    if v3 < 0:
        # ximax=v1-v2/v3
        ximax = v1 - v2 / (v3 - minxi)
        for i in range(len(y)):
            if y[i] > ximax:
                y[i] = ximax
    
    if v3 > 0:
        # ximin=v1-v2/v3
        ximin = v1 - v2 / (v3 + minxi)
        for i in range(len(y)):
            if y[i] < ximin:
                y[i] = ximin
    
    return y

#  Deal with situations in which the user wants d or p outside the GPD range
def fixgpdrange(y, v1, v2, v3):
    y = np.array(y, copy=True)
    
    if v3 < 0:
        ximin = v1
        ximax = v1 - v2 / v3
        for i in range(len(y)):
            if y[i] < ximin:
                y[i] = ximin
            if y[i] > ximax:
                y[i] = ximax
    else:
        for i in range(len(y)):
            if y[i] < v1:
                y[i] = v1
    
    return y


def movexiawayfromzero(xi):
    #Move xi away from zero a bit
    minxi = 10**-7
    if abs(xi) < minxi:
        if xi >= 0:
            # the problem with this idea could be that increasing xi, when xi is positive,
            # increases the minimum of the gev, and some y values might then lie below the minimum
            # but I've made a change to the 'fixgevrange' routine to deal with that
            # and that seems to fix the problem now
            xi = minxi
        else:
            xi = -minxi
    return xi


def to_array(x):
    # so that certain functions can take either a single number, or an array
    if (isinstance(x, float) or isinstance(x, int)):
        x = [x]
    x = np.asarray(x)
    return x


def ru(arg, **kwargs):
    raise Exception('rust not yet implemented')


def maket0(t0, n0, t):
    # if t0 is specified, does nothing
    # if t0 isn't specified, calculates t0 from n0
    
    if (t0 is None or np.isnan(t0)) and (n0 is None or np.isnan(n0)):
        raise Exception("Either t0 or n0 must be specified")
    if (t0 is not None and not np.isnan(t0)) and (n0 is not None and not np.isnan(n0)):
        raise Exception("Only one of t0 or n0 must be specified")
    if t0 is None or np.isnan(t0):
        t0 = t[n0 - 1]  # Convert to 0-based indexing
    
    return t0

def makemuhat0(t0, n0, t, mle_params):
    t = np.array(t)
    mle_params = np.array(mle_params)
    
    muhat = mle_params[0] + mle_params[1] * t
    
    if t0 is None or np.isnan(t0):
        muhat0 = muhat[n0 - 1]  # Convert to 0-based indexing
    else:
        muhat0 = mle_params[0] + mle_params[1] * t0
    
    return muhat0

def maketa0(t0, n0, t):
    t = np.array(t)
    ta = t - np.mean(t)
    
    if t0 is None or np.isnan(t0):
        ta0 = ta[n0 - 1]  # Convert to 0-based indexing
    else:
        ta0 = t0 - np.mean(t)
    
    return ta0

# make Standard Errors from lddi
def make_se(nx, lddi):
    nd = lddi.shape[0]
    standard_errors = np.zeros(nd, dtype=object)
    
    for i in range(nd):
        if lddi[i, i] > 0:
            standard_errors[i] = "square root not possible"
        else:
            standard_errors[i] = np.sqrt(-lddi[i, i] / nx)
    
    return standard_errors

def make_waic(x, fhatx, lddi, lddd, f1f, lambdad, f2f, dim):
    # waic seems to be a bust because it doesn't make sense for the mle models...it doesn't penalize them at all for parameters
    # the variance version always penalizes twice as much as the difference version...is there a factor of 2 wrong somewhere?
    
    # x is implicit in f1f and f2f
    
    nx = len(x)
    lfhatx = np.log(fhatx)
    fhatxsq = fhatx * fhatx
    lfhatxof = lfhatx / fhatx
    omlfhatx = 1 - lfhatx
    
    outerf1f = np.zeros((dim, dim, nx))
    for i in range(nx):
        for j in range(dim):
            for k in range(dim):
                outerf1f[j, k, i] = np.outer(f1f[j, i], f1f[k, i])
    
    # first term is the llpd
    # which is just the in-sample log-likelihood
    # for mle, should match the log-likelihood at the max
    # calc the probabilities, take the log, add them together
    t1 = np.zeros((dim, nx))
    t2 = np.zeros((dim, dim, nx))
    for i in range(nx):
        t1[:, i] = f1f[:, i]
        t2[:, :, i] = f2f[:, :, i]
    
    dq = cp_dmgs.dmgs(lddi, lddd, t1, lambdad, t2, dim=dim) / nx
    dqq = np.maximum(fhatx + dq, sys.float_info.epsilon)
    logf1_rhp = np.log(dqq)
    sumlogf1_rhp = np.sum(logf1_rhp)
    lppd_rhp = sumlogf1_rhp
    
    # second term is the penalty term
    # this version of it is a difference
    # seems to be the same for maxlik
    t1 = np.zeros((dim, nx))
    t2 = np.zeros((dim, dim, nx))
    for i in range(nx):
        t1[:, i] = f1f[:, i] / fhatx[i]
        t2[:, :, i] = f2f[:, :, i] / fhatx[i] - outerf1f[:, :, i] / fhatxsq[i]
    
    logf2_rhp = lfhatx + cp_dmgs.dmgs(lddi, lddd, t1, lambdad, t2, dim=dim) / nx
    sumlogf2_rhp = np.sum(logf2_rhp)
    pwaic1_rhp = 2 * (sumlogf1_rhp - sumlogf2_rhp)
    
    # alternative version uses the variance thing for the second term
    
    # this version of it is a variance
    # it penalizes more...is there really a factor of 2 needed
    t2inner = np.zeros((dim, dim, nx))
    for i in range(nx):
        t1[:, i] = 2 * lfhatxof[i] * f1f[:, i]
        t2inner[:, :, i] = f2f[:, :, i] * lfhatx[i] + outerf1f[:, :, i] * omlfhatx[i] / fhatx[i]
        t2[:, :, i] = 2 * t2inner[:, :, i] / fhatx[i]
    
    dq = cp_dmgs.dmgs(lddi, lddd, t1, lambdad, t2, dim=dim)
    sumvlogf_rhp = np.sum(lfhatx * lfhatx + dq / nx - logf2_rhp * logf2_rhp)
    pwaic2_rhp = sumvlogf_rhp
    
    waic1 = np.zeros(3)
    waic2 = np.zeros(3)
    waic1[0] = lppd_rhp
    waic1[1] = -pwaic1_rhp
    waic1[2] = waic1[0] + waic1[1]
    waic2[0] = lppd_rhp
    waic2[1] = -pwaic2_rhp
    waic2[2] = waic2[0] + waic2[1]
    
    return {'waic1': waic1, 'waic2': waic2}

def make_cwaic(x, fhatx, lddi, lddd, f1f, lambdad, f2f, dim):
    # waic seems to be a bust because it doesn't make sense for the mle models...it doesn't penalize them at all for parameters
    # the variance version always penalizes twice as much as the difference version...is there a factor of 2 wrong somewhere?
    
    # x is implicit in f1f and f2f
    
    nx = len(x)
    lfhatx = np.log(fhatx)
    fhatxsq = fhatx * fhatx
    lfhatxof = lfhatx / fhatx
    omlfhatx = 1 - lfhatx
    
    outerf1f = np.zeros((dim, dim, nx))
    for i in range(nx):
        for j in range(dim):
            for k in range(dim):
                outerf1f[j, k, i] = np.outer(f1f[j, i], f1f[k, i])
    
    # first term is the llpd
    # which is just the in-sample log-likelihood
    # for mle, should match the log-likelihood at the max
    # calc the probabilities, take the log, add them together
    t1 = np.zeros((dim, nx))
    t2 = np.zeros((dim, dim, nx))
    for i in range(nx):
        t1[:, i] = f1f[:, i]
        t2[:, :, i] = f2f[:, :, i]
    
    dq = cp_dmgs.dmgs(lddi, lddd, t1, lambdad, t2, dim=dim) / nx
    dqq = np.maximum(fhatx + dq, sys.float_info.epsilon)
    logf1_rhp = np.log(dqq)
    sumlogf1_rhp = np.sum(logf1_rhp)
    lppd_rhp = sumlogf1_rhp
    
    # second term is the penalty term
    # this version of it is a difference
    # seems to be the same for maxlik
    t1 = np.zeros((dim, nx))
    t2 = np.zeros((dim, dim, nx))
    for i in range(nx):
        t1[:, i] = f1f[:, i] / fhatx[i]
        t2[:, :, i] = f2f[:, :, i] / fhatx[i] - outerf1f[:, :, i] / fhatxsq[i]
    
    logf2_rhp = lfhatx + cp_dmgs.dmgs(lddi, lddd, t1, lambdad, t2, dim=dim) / nx
    sumlogf2_rhp = np.sum(logf2_rhp)
    pwaic1_rhp = 2 * (sumlogf1_rhp - sumlogf2_rhp)
    
    # alternative version uses the variance thing for the second term
    
    # this version of it is a variance
    # it penalizes more...is there really a factor of 2 needed
    t2inner = np.zeros((dim, dim, nx))
    for i in range(nx):
        t1[:, i] = 2 * lfhatxof[i] * f1f[:, i]
        t2inner[:, :, i] = f2f[:, :, i] * lfhatx[i] + outerf1f[:, :, i] * omlfhatx[i] / fhatx[i]
        t2[:, :, i] = 2 * t2inner[:, :, i] / fhatx[i]
    
    dq = cp_dmgs.dmgs(lddi, lddd, t1, lambdad, t2, dim=dim)
    sumvlogf_rhp = np.sum(lfhatx * lfhatx + dq / nx - logf2_rhp * logf2_rhp)
    pwaic2_rhp = sumvlogf_rhp
    
    waic1 = np.zeros(3)
    waic2 = np.zeros(3)
    waic1[0] = lppd_rhp
    waic1[1] = -pwaic1_rhp
    waic1[2] = waic1[0] + waic1[1]
    waic2[0] = lppd_rhp
    waic2[1] = -pwaic2_rhp
    waic2[2] = waic2[0] + waic2[1]
    
    return {'waic1': waic1, 'waic2': waic2}

def make_maic(ml_value, nparams):
    """
    Calculate MAIC
    Returns Vector of 3 values - the two components of MAIC, and their sum
    """
    maic = np.zeros(3)
    maic[0] = ml_value
    maic[1] = -nparams
    maic[2] = maic[0] + maic[1]
    return maic

def rust_pumethod():
    method = ("The cp results are based posterior simulation using "
              "ratio of uniforms sampling (RUST), "
              "using the predictive matching right Haar prior.")
    method = textwrap.fill(method, width=10000)
    
    return method

def analytic_cpmethod():
    method = ("The cp results are based on an analytic solution of "
              "the Bayesian prediction integral, "
              "using the predictive matching right Haar prior.")
    method = textwrap.fill(method, width=10000)
    
    return method

def rhp_dmgs_cpmethod():
    method = ("The cp results are based on the DMGS approximation of "
              "the Bayesian prediction integral, "
              "using the predictive matching right Haar prior.")
    method = textwrap.fill(method, width=10000)
    
    return method

def crhpflat_dmgs_cpmethod():
    method = ("The cp results are based on the DMGS approximation of "
              "the Bayesian prediction integral, "
              "using the CRHP/flat prior.")
    method = textwrap.fill(method, width=10000)
    
    return method

def adhoc_dmgs_cpmethod():
    method = ("The cp results are based on the DMGS approximation of "
              "the Bayesian prediction integral, "
              "using an ad-hoc prior.")
    method = textwrap.fill(method, width=10000)
    
    return method

def makeq(yy, pp):
    """Calculates quantiles from simulations by inverting the Hazen CDF"""
    yy = np.array(yy)
    pp = np.array(pp)
    nyy = len(yy)
    npp = len(pp)
    syy = np.sort(yy)
    lambda_vals = nyy * pp + 0.5
    qq = np.zeros(npp)
    aa = np.floor(lambda_vals).astype(int)
    eps = lambda_vals - aa
    
    for ii in range(npp):
        if lambda_vals[ii] <= 1:
            qq[ii] = syy[0]
        elif lambda_vals[ii] >= nyy:
            qq[ii] = syy[nyy - 1]
        else:
            qq[ii] = (1 - eps[ii]) * syy[aa[ii] - 1] + eps[ii] * syy[aa[ii]]
    
    return qq

def nopdfcdfmsg():
    """
    Message to explain why GEV and GPD d*** and p*** routines
    don't return DMGS pdfs and cdfs
    """
    msg = "For the pdf and cdf for the " \
    " and GPD, considered as a function of the rv, DMGS doesn't return anything."
    msg = msg + " That's because DMGS can't work outside the range limits."
    msg = msg + " Instead, you could use:"
    msg = msg + " (a) q***, for the inverse CDF, or"
    msg = msg + " (b) q***, with pdf=TRUE, which returns the pdf as a function of probability, or"
    msg = msg + " (c) d*** or p***, with rust=TRUE, which uses rust to generate the pdf and cdf."
    
    return msg


import sympy
import autodiff

x, v1, v2 = sympy.symbols('x v1 v2')

f = (v1/v2)*((x/v2)**(v1-1))*sympy.exp(-((x/v2)**v1))
p = 1-sympy.exp(-((x/v2)**v1))
logf = sympy.log(v1)-sympy.log(v2)+(v1-1)*sympy.log(x/v2)-(x/v2)**v1
                                 
with open('output_weibull.py', 'w') as file:
    file.write('import numpy as np\n')
    file.write(autodiff.diff_to_code(f, [v1, v2], 'weibull_fd', 1))
    file.write(autodiff.diff_to_code(f, [v1, v2], 'weibull_fdd', 2))
    file.write(autodiff.diff_to_code(p, [v1, v2], 'weibull_pd', 1))
    file.write(autodiff.diff_to_code(p, [v1, v2], 'weibull_pdd', 2))
    file.write(autodiff.diff_to_code(logf, [v1, v2], 'weibull_logfdd', 2))
    file.write(autodiff.diff_to_code(logf, [v1, v2], 'weibull_logfddd', 3))

file.close()


import numpy as np
import sys
import scipy.stats
import scipy.optimize

from . import gumbel_derivs
from . import utils as cp_utils
from . import evaluate_dmgs_equation as cp_dmgs


def gumbel_waic(waicscores, x, v1hat, v2hat, lddi, lddd, lambdad):
    if waicscores:
        f1f = gumbel_derivs.gumbel_f1fa(x, v1hat, v2hat)
        f2f = gumbel_derivs.gumbel_f2fa(x, v1hat, v2hat)
        
        fhatx = scipy.stats.gumbel_r.pdf(x, loc=v1hat, scale=v2hat)
        waic = cp_utils.make_waic(x, fhatx, lddi, lddd, f1f, lambdad, f2f, dim=2)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
    else:
        waic1 = "waicscores not selected"
        waic2 = "waicscores not selected"
    
    return {'waic1': waic1, 'waic2': waic2}

def gumbel_logf(params, x):
    m = params[0]
    s = np.maximum(params[1], sys.float_info.epsilon)
    logf = np.sum(scipy.stats.gumbel_r.logpdf(x, loc=m, scale=s)) - np.log(s)
    return logf

def gumbel_loglik(vv, x):
    loglik = np.sum(scipy.stats.gumbel_r.logpdf(x, loc=vv[0], scale=max(vv[1], sys.float_info.epsilon)))
    return loglik

def gumbel_means(means, ml_params, lddi, lddd, lambdad_rhp, nx, dim=2):
    """MLE and RHP predictive means """
    if means:
        # intro
        eulerconstant = 0.57721566490153286060651209008240243104215933593992
        
        # ml mean
        ml_mean = ml_params[0] + ml_params[1] * eulerconstant
        
        # rhp mean
        meand1 = np.zeros((2, 1))
        meand1[0, 0] = 1
        meand1[1, 0] = eulerconstant
        meand2 = np.zeros((2, 2, 1))  # but all zero for gumbel
        dmean = cp_dmgs.dmgs(lddi, lddd, meand1, lambdad_rhp, meand2, dim=2)
        rh_mean = ml_mean + dmean / nx
    else:
        ml_mean = "means not selected"
        rh_mean = "means not selected"
    
    return {'ml_mean': ml_mean, 'rh_mean': rh_mean}

def gumbel_logscores(logscores, x):
    """Log scores for MLE and RHP predictions calculated using leave-one-out """
    if logscores:
        nx = len(x)
        ml_oos_logscore = 0
        rh_oos_logscore = 0
        for i in range(nx):
            x1 = np.concatenate([x[:i], x[i+1:]])
            
            dd = dgumbelsub(x1, x[i])
            
            ml_params = dd['ml_params']
            
            ml_pdf = dd['ml_pdf']
            ml_oos_logscore = ml_oos_logscore + np.log(ml_pdf)
            
            rh_pdf = dd['rh_pdf']
            rh_oos_logscore = rh_oos_logscore + np.log(max(rh_pdf, sys.float_info.epsilon))
    else:
        ml_oos_logscore = "logscores not selected"
        rh_oos_logscore = "logscores not selected"
    
    return {'ml_oos_logscore': ml_oos_logscore, 'rh_oos_logscore': rh_oos_logscore}

def dgumbelsub(x, y):
    """Densities from MLE and RHP"""
    nx = len(x)
    
    v1start = np.mean(x)
    v2start = np.std(x)
    
    # Use minimize with negative log-likelihood since optim with fnscale=-1 maximizes
    def neg_loglik(params):
        return -gumbel_loglik(params, x)
    
    opt = scipy.optimize.minimize(neg_loglik, [v1start, v2start])
    v1hat = opt.x[0]
    v2hat = opt.x[1]
    ml_params = np.array([v1hat, v2hat])
    
    # mle
    ml_pdf = scipy.stats.gumbel_r.pdf(y, loc=v1hat, scale=v2hat)
    ml_cdf = scipy.stats.gumbel_r.cdf(y, loc=v1hat, scale=v2hat)
    
    # rhp
    ldd = gumbel_derivs.gumbel_ldda(x, v1hat, v2hat)
    lddi = np.linalg.inv(ldd)
    lddd = gumbel_derivs.gumbel_lddda(x, v1hat, v2hat)
    
    f1 = gumbel_derivs.gumbel_f1fa(y, v1hat, v2hat)
    f2 = gumbel_derivs.gumbel_f2fa(y, v1hat, v2hat)
    
    p1 = gumbel_derivs.gumbel_p1fa(y, v1hat, v2hat)
    p2 = gumbel_derivs.gumbel_p2fa(y, v1hat, v2hat)
    
    lambdad_rhp = np.array([0, -1/v2hat])
    df1 = cp_dmgs.dmgs(lddi, lddd, f1, lambdad_rhp, f2, dim=2)
    dp1 = cp_dmgs.dmgs(lddi, lddd, p1, lambdad_rhp, p2, dim=2)
    rh_pdf = np.maximum(ml_pdf + df1/nx, 0)
    rh_cdf = np.minimum(np.maximum(ml_cdf + dp1/nx, 0), 1)
    
    # return
    return {
        'ml_params': ml_params,
        'ml_pdf': ml_pdf,
        'rh_pdf': rh_pdf,
        'ml_cdf': ml_cdf,
        'rh_cdf': rh_cdf
    }



import numpy as np
import scipy.optimize as optimize
import scipy.special as special
from scipy.stats import genextreme

from . import utils as cp_utils
from . import genextreme_derivs as cp_gev_c


def rgev_minmax(nx, mu, sigma, xi, minxi=-1, maxxi=1):
    """rgev but with maxlik xi guaranteed within bounds"""
    xihat = -9999
    while (xihat < minxi) or (xihat > maxxi):
        xx = genextreme.rvs(c=-xi, loc=mu, scale=sigma, size=nx)
        ics = gev_setics(xx, np.array([0, 0, 0]))
        opt1 = optimize.minimize(lambda params: -gev_loglik(params, xx), 
                               ics, method='BFGS')
        xihat = opt1.x[2]
    return xx


def gev_waic(waicscores, x, v1hat, v2hat, v3hat, lddi, lddd, lambdad):
    #Returns Dictionary containing waic1 and waic2
    if waicscores:
        f1f = cp_gev_c.gev_f1fa(x, v1hat, v2hat, v3hat)
        f2f = cp_gev_c.gev_f2fa(x, v1hat, v2hat, v3hat)
        
        fhatx = genextreme.pdf(x, c=-v3hat, loc=v1hat, scale=v2hat)
        waic = cp_utils.make_waic(x, fhatx, lddi, lddd, f1f, lambdad, f2f, dim=3)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
    else:
        waic1 = "waicscores not selected"
        waic2 = "waicscores not selected"
    return {'waic1': waic1, 'waic2': waic2}


def gev_logf(params, x):
    #Returns   (float) Log likelihood value
    mu = params[0]
    sc = np.maximum(params[1], np.finfo(float).eps)
    sh = params[2]
    
    logf = np.sum(genextreme.logpdf(x, c=-sh, loc=mu, scale=sc)) - np.log(sc)
    return logf


def gev_setics(x, ics):
    # set initial conditions
    if (ics[0] == 0) and (ics[1] == 0) and (ics[2] == 0):
        ics[0] = np.mean(x)
        ics[1] = np.std(x, ddof=1)
        ics[2] = 0
    return ics


def gev_pwm_params(x):
    """
    Parameters
        x : numpy.ndarray
    Returns 
        array: PWM parameter estimates
    """
    raise Exception('The PWM method is not yet implemented in fitdistcp; please use Maximum Likelihood.')
    # Note: fExtremes::gevFit with PWM method is not directly available in Python
    # This would need a custom PWM implementation or alternative library
    # Simple method of moments approximation as placeholder
    pw_params = np.zeros(3)
    pw_params[0] = np.mean(x)           # mu
    pw_params[1] = np.std(x, ddof=1)    # sigma
    pw_params[2] = 0                    # xi
    
    return pw_params


def gev_loglik(vv, x):
    # return log-likelihood gev
    loglik = np.sum(genextreme.logpdf(x, c=-vv[2], loc=vv[0], 
                                     scale=np.maximum(vv[1], np.finfo(float).eps)))
    return loglik


def gev_means(means, ml_params, lddi, lddd, lambdad_rh_flat, lambdad_custom, nx, dim=3):
    """
    Analytical Expressions for Predictive Means
    RHP mean based on the expectation of DMGS equation 2.1
    Returns Dictionary containing ml_mean, rh_flat_mean, custom_mean
    """
    if means:
        # intro
        eulerconstant = 0.57721566490153286060651209008240243104215933593992
        mu = ml_params[0]
        sigma = ml_params[1]
        xi = ml_params[2]
        
        # set up derivative arrays
        meand1 = np.zeros((3, 1))
        meand2 = np.zeros((3, 3, 1))  # but all zero for gumbel
        
        if ml_params[2] == 0:
            # xi=0 case
            
            # mle
            ml_mean = mu + sigma * eulerconstant
            # calculate first derivative array for bayesian xi=0 cases
            meand1[0, 0] = 1
            meand1[1, 0] = eulerconstant
            # meand2 is all zero as initialized
        else:
            # non-gumbel case
            g0 = special.gamma(1 - xi)
            g1 = g0 * special.digamma(1 - xi)
            g2 = (special.polygamma(1, 1 - xi) * g0 * g0 + g1 * g1) / g0
            
            # mle
            ml_mean = mu + sigma * (g0 - 1) / xi
            # calculate first derivative array for bayesian xi!=0 cases
            meand1[0, 0] = 1
            meand1[1, 0] = (g0 - 1) / xi
            meand1[2, 0] = (1 - g0 - xi * g1) * sigma / (xi * xi)
            # calculate second derivative array (only has 1 non-zero term!)
            meand2[1, 2, 0] = (1 - g0 - xi * g1) / (xi * xi)
            meand2[2, 1, 0] = meand2[1, 2, 0]
            meand2[2, 2, 0] = sigma * (-2 + 2 * g0 + 2 * xi * g1 + xi * xi * g2) / (xi * xi * xi)
        
        # I've realized now that when we integrate over xi, the mean is Inf
        rh_flat_mean = np.inf
        custom_mean = np.inf
    else:
        ml_mean = "means not selected"
        rh_flat_mean = "means not selected"
        custom_mean = "means not selected"
    
    # return
    return {
        'ml_mean': ml_mean,
        'rh_flat_mean': rh_flat_mean,
        'custom_mean': custom_mean
    }


def dgevsub(x, y, ics):
    """
    Densities for 5 predictions
    Returns Dictionary containing ml_params, ml_pdf, ml_cdf
    """
    
    ics = gev_setics(x, ics)
    opt = optimize.minimize(lambda params: -gev_loglik(params, x), 
                           ics, method='BFGS')
    v1hat = opt.x[0]
    v2hat = opt.x[1]
    v3hat = opt.x[2]
    ml_params = np.array([v1hat, v2hat, v3hat])
    
    # mle
    ml_pdf = genextreme.pdf(y, c=-v3hat, loc=v1hat, scale=v2hat)
    ml_cdf = genextreme.cdf(y, c=-v3hat, loc=v1hat, scale=v2hat)
    
    # return
    return {
        'ml_params': ml_params,
        'ml_pdf': ml_pdf,
        'ml_cdf': ml_cdf
    }


import numpy as np
import sys
import scipy.stats

from . import expon_derivs as exp_derivs
from . import utils as cp_utils


def exp_waic(waicscores, x, v1hat):
    """
    Waicscores
    @inherit manwaic return
    @inheritParams manf
    """
    if waicscores:
        f1f = exp_derivs.exp_f1fa(x, v1hat)
        f2f = exp_derivs.exp_f2fa(x, v1hat)
        
        ldd = exp_derivs.exp_ldda(x, v1hat)
        lddi = np.linalg.inv(ldd)
        lddd = exp_derivs.exp_lddda(x, v1hat)
        fhatx = scipy.stats.expon.pdf(x, scale=1/v1hat)
        lambdad_rhp = -1/v1hat
        waic = cp_utils.make_waic(x, fhatx, lddi, lddd, f1f, lambdad_rhp, f2f, dim=1)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
    else:
        waic1 = "waicscores not selected"
        waic2 = "waicscores not selected"
    
    return {'waic1': waic1, 'waic2': waic2}

def exp_logf(params, x):
    """
    Logf for RUST
    @inherit manlogf return
    @inheritParams manf
    """
    l = np.maximum(params[0], sys.float_info.epsilon)
    logf = np.sum(scipy.stats.expon.logpdf(x, scale=1/l)) - np.log(l)
    return logf

def exp_logscores(logscores, x):
    """
    Log scores for MLE and RHP predictions calculated using leave-one-out
    @inherit manlogscores return
    @inheritParams manf
    """
    if logscores:
        # I could put the logs inside dexpsub, but I'd have to actually calculate the log for the rhp case
        
        nx = len(x)
        ml_oos_logscore = 0
        rh_oos_logscore = 0
        for i in range(nx):
            x1 = np.concatenate([x[:i], x[i+1:]])
            dd = dexpsub(x1, x[i])
            ml_params1 = (nx-1) / np.sum(x1)
            # ml
            ml_pdf = dd['ml_pdf']
            ml_oos_logscore = ml_oos_logscore + np.log(ml_pdf)
            # rhp
            rh_pdf = dd['rh_pdf']
            rh_oos_logscore = rh_oos_logscore + np.log(rh_pdf)
    else:
        ml_oos_logscore = "logscores not selected"
        rh_oos_logscore = "logscores not selected"
    
    return {'ml_oos_logscore': ml_oos_logscore, 'rh_oos_logscore': rh_oos_logscore}

def dexpsub(x, y):
    """
    Densities from MLE and RHP
    @inherit mandsub return
    @inheritParams manf
    """
    nx = len(x)
    
    # ml
    ml_params = nx / np.sum(x)
    ml_pdf = scipy.stats.expon.pdf(y, scale=1/ml_params)
    ml_cdf = scipy.stats.expon.cdf(y, scale=1/ml_params)
    
    # rhp pdf
    sx = np.sum(x)
    top = sx**nx
    bot1 = (sx + y)**(nx + 1)
    rh_pdf = nx * top / bot1
    rh_pdf = np.maximum(rh_pdf, 0)
    
    # rhp cdf
    bot2 = (sx + y)**nx
    rh_cdf = 1 - top / bot2
    rh_cdf = np.minimum(np.maximum(rh_cdf, 0), 1)
    
    # return
    return {
        'ml_params': ml_params,
        'ml_pdf': ml_pdf,
        'rh_pdf': rh_pdf,
        'ml_cdf': ml_cdf,
        'rh_cdf': rh_cdf
    }




import numpy as np
import json
import warnings
import matplotlib.pyplot as plt

from scipy.optimize import minimize
from scipy.stats import expon
from scipy.stats import genextreme
from scipy.stats import genpareto
from scipy.stats import gumbel_r

from . import gumbel as cp_gumbel
from . import gumbel_libs as cp_gumbel_libs
from . import expon as cp_expon
from . import expon_libs as cp_expon_libs
from . import genpareto as cp_genpareto
from . import genpareto_libs as cp_genpareto_libs
from . import genextreme as cp_genextreme
from . import genextreme_libs as cp_genextreme_libs

from . import reltest_libs


warnings.filterwarnings("ignore", category=RuntimeWarning)


def reltest_expon(file, ntrials=100, nx=30, p=0.0001*np.asarray(range(1,10000))):
    '''
    Reliability test.

    Parameters
    ----------
    desired_p : array_like
        Probabilities at which to calculate quantiles.
    ntrials : int
        Number of trials to average over.
    nx : int
        Number of samples per trial.
    
    Returns
    -------
    dict
        Dictionary with keys:
            'actual_p_ml' : array_like
                Achieved probabilities using ML quantiles.
            'actual_p_cp' : array_like
                Achieved probabilities using CP quantiles.

    Each trial generates nx samples and calculates quantiles using the two methods.
    The difference between the methods is clearest when nx is in the range of 20-60.
    Increasing ntrials reduces the effect of random variations in the trials (100 is sufficient for many purposes).
    '''

    scale = 1

    p_actual_ml_total = np.zeros(len(p))
    p_actual_cp_total = np.zeros(len(p))

    for i in range(ntrials):
        x = expon.rvs(scale=scale, size=nx)
        
        # calculate quantile using ML and CP methods
        # ML:
        ics = [scale]
        scale_ml = minimize(lambda params: -cp_expon_libs.exp_logf(params, x), ics, method='BFGS').x 
        q_ml = expon.ppf(p, scale=scale_ml)

        # CP:
        q_cp = cp_expon.ppf(x, p)['cp_quantiles']

        # feed back in for the actual probability
        p_actual_ml_total += expon.cdf(q_ml, scale=scale)
        p_actual_cp_total += expon.cdf(q_cp, scale=scale)

    p_actual_ml_avg = p_actual_ml_total / ntrials
    p_actual_cp_avg = p_actual_cp_total / ntrials

    result = {
        'actual_p_ml' : np.ndarray.tolist(p_actual_ml_avg), 
        'actual_p_cp': np.ndarray.tolist(p_actual_cp_avg), 
        'p': np.ndarray.tolist(p)
        }

    json.dump(result, file)




def reltest_genextreme(file, ntrials=100, nx=30, p=0.0001*np.asarray(range(1,10000))):
    '''
    Reliability test.

    Parameters
    ----------
    desired_p : array_like
        Probabilities at which to calculate quantiles.
    ntrials : int
        Number of trials to average over.
    nx : int
        Number of samples per trial.
    
    Returns
    -------
    dict
        Dictionary with keys:
            'actual_p_ml' : array_like
                Achieved probabilities using ML quantiles.
            'actual_p_cp' : array_like
                Achieved probabilities using CP quantiles.

    Each trial generates nx samples and calculates quantiles using the two methods.
    The difference between the methods is clearest when nx is in the range of 20-60.
    Increasing ntrials reduces the effect of random variations in the trials (100 is sufficient for many purposes).
    '''
    xi = 0
    mu = 0
    sigma = 1

    p_actual_ml_total = np.zeros(len(p))
    p_actual_cp_total = np.zeros(len(p))

    for i in range(ntrials):
        x = genextreme.rvs(-xi, loc=mu, scale=sigma, size=nx)
        
        # calculate quantile using ML and CP methods
        # ML:
        ics = [mu, sigma, xi]
        mu_ml, sigma_ml, xi_ml = minimize(lambda params: -cp_genextreme_libs.gev_loglik(params, x), ics, method='BFGS').x
        q_ml = genextreme.ppf(p, -xi_ml, loc=mu_ml, scale=sigma_ml) 

        # CP:
        q_cp = cp_genextreme.ppf(x, p)['cp_quantiles']   

        # feed back in for the actual probability
        p_actual_ml_total += genextreme.cdf(q_ml, -xi, loc=mu, scale=sigma)
        p_actual_cp_total += genextreme.cdf(q_cp, -xi, loc=mu, scale=sigma)

    p_actual_ml_avg = p_actual_ml_total / ntrials
    p_actual_cp_avg = p_actual_cp_total / ntrials

    result = {
        'actual_p_ml' : np.ndarray.tolist(p_actual_ml_avg), 
        'actual_p_cp': np.ndarray.tolist(p_actual_cp_avg), 
        'p': np.ndarray.tolist(p)
        }

    json.dump(result, file)


def reltest_genpareto(file, ntrials=100, nx=30, p=0.0001*np.asarray(range(1,10000))):
    '''
    Reliability test.

    Parameters
    ----------
    desired_p : array_like
        Probabilities at which to calculate quantiles.
    ntrials : int
        Number of trials to average over.
    nx : int
        Number of samples per trial.
    
    Returns
    -------
    dict
        Dictionary with keys:
            'actual_p_ml' : array_like
                Achieved probabilities using ML quantiles.
            'actual_p_cp' : array_like
                Achieved probabilities using CP quantiles.

    Each trial generates nx samples and calculates quantiles using the two methods.
    The difference between the methods is clearest when nx is in the range of 20-60.
    Increasing ntrials reduces the effect of random variations in the trials (100 is sufficient for many purposes).
    '''
    kloc = 0
    sigma = 1
    xi = 0

    p_actual_ml_total = np.zeros(len(p))
    p_actual_cp_total = np.zeros(len(p))

    for i in range(ntrials):
        x = genpareto.rvs(-xi, loc=kloc, scale=sigma, size=nx)
        
        # calculate quantile using ML and CP methods
        # ML:
        ics = [sigma, xi]
        sigma_ml, xi_ml = minimize(lambda params: -cp_genpareto_libs.gpd_k1_loglik(params, x, kloc), ics, method='BFGS').x
        q_ml = genpareto.ppf(p, -xi_ml, loc=kloc, scale=sigma_ml)

        # CP:
        q_cp = cp_genpareto.ppf(x, p)['cp_quantiles']

        # feed back in for the actual probability
        p_actual_ml_total += genpareto.cdf(q_ml, -xi, loc=kloc, scale=sigma)
        p_actual_cp_total += genpareto.cdf(q_cp, -xi, loc=kloc, scale=sigma)

    p_actual_ml_avg = p_actual_ml_total / ntrials
    p_actual_cp_avg = p_actual_cp_total / ntrials

    result = {
        'actual_p_ml' : np.ndarray.tolist(p_actual_ml_avg), 
        'actual_p_cp': np.ndarray.tolist(p_actual_cp_avg), 
        'p': np.ndarray.tolist(p)
        }

    json.dump(result, file)


def reltest_gumbel(file, ntrials=100, nx=30, p=0.0001*np.asarray(range(1,10000))):
    '''
    Reliability test.

    Parameters
    ----------
    desired_p : array_like
        Probabilities at which to calculate quantiles.
    ntrials : int
        Number of trials to average over.
    nx : int
        Number of samples per trial.
    
    Returns
    -------
    dict
        Dictionary with keys:
            'actual_p_ml' : array_like
                Achieved probabilities using ML quantiles.
            'actual_p_cp' : array_like
                Achieved probabilities using CP quantiles.

    Each trial generates nx samples and calculates quantiles using the two methods.
    The difference between the methods is clearest when nx is in the range of 20-60.
    Increasing ntrials reduces the effect of random variations in the trials (100 is sufficient for many purposes).
    '''
    loc = 0
    scale = 1

    p_actual_ml_total = np.zeros(len(p))
    p_actual_cp_total = np.zeros(len(p))

    for i in range(ntrials):
        x = gumbel_r.rvs(loc=loc, scale=scale, size=nx)
        
        # calculate quantile using ML and CP methods
        # ML:
        ics = [loc, scale]
        loc_ml, scale_ml = minimize(lambda params: -cp_gumbel_libs.gumbel_logf(params, x), ics, method='BFGS').x 
        q_ml = gumbel_r.ppf(p, loc=loc_ml, scale=scale_ml)

        # CP:
        q_cp = cp_gumbel.ppf(x, p)['cp_quantiles']

        # feed back in for the actual probability
        p_actual_ml_total += gumbel_r.cdf(q_ml, loc=loc, scale=scale)
        p_actual_cp_total += gumbel_r.cdf(q_cp, loc=loc, scale=scale)

    p_actual_ml_avg = p_actual_ml_total / ntrials
    p_actual_cp_avg = p_actual_cp_total / ntrials

    result = {
        'actual_p_ml' : np.ndarray.tolist(p_actual_ml_avg), 
        'actual_p_cp': np.ndarray.tolist(p_actual_cp_avg), 
        'p': np.ndarray.tolist(p)
        }

    json.dump(result, file)


# -----------------------------------------------------------------------------------------


def single_plot(data, ax):
    ax.plot(data['x'], data['y_ml'],  label='ML', color='red', linewidth=1)
    ax.plot(data['x'], data['y_cp'],  label='CP', color='blue')
    ax.plot(data['benchmark']['x'], data['benchmark']['y'], color='black', label='Benchmark')
    ax.set_xlim(data['limits']['x'])
    ax.set_ylim(data['limits']['y'])
    ax.set_xlabel(data['xlabel'])
    ax.set_ylabel(data['ylabel'])
    ax.set_title(data['title'])
    ax.legend()                         # crowds the graphs somewhat   



def plot(file):
    '''Passed the open file containing output from reltests, generates 5 graphs to compare ML and CP results.'''
    result = json.load(file)
    actual_p_ml = result['actual_p_ml']
    actual_p_cp = result['actual_p_cp']
    p = result['p']

    data = [
        reltest_libs.format_data_a(actual_p_ml, actual_p_cp, p), 
        reltest_libs.format_data_b(actual_p_ml, actual_p_cp, p), 
        reltest_libs.format_data_d(actual_p_ml, actual_p_cp, p), 
        reltest_libs.format_data_h(actual_p_ml, actual_p_cp, p), 
        reltest_libs.format_data_i(actual_p_ml, actual_p_cp, p)
        ]
    fig, axs = plt.subplots(3,2)
    single_plot(data[0], axs[0,0])
    single_plot(data[1], axs[0,1])
    single_plot(data[2], axs[1,0])
    single_plot(data[3], axs[1,1])
    single_plot(data[4], axs[2,0])
    fig.delaxes(axs[2,1])
    fig.tight_layout()
    plt.show()



import scipy
import sympy
import numpy as np
import autodiff

x, v1, v2 = sympy.symbols('x v1 v2')

f = (v2**(-v1))*(1/sympy.gamma(v1))*(x**(v1-1))*sympy.exp(-(x/v2))
logf = -v1*sympy.log(v2)-sympy.log(sympy.gamma(v1))+(v1-1)*sympy.log(x)-x/v2
                                 
with open('output_gamma.py', 'w') as file:
    file.write('import numpy as np\n')
    file.write(autodiff.diff_to_code(f, [v1, v2], 'gamma_fd', 1))
    file.write(autodiff.diff_to_code(f, [v1, v2], 'gamma_fdd', 2))
    file.write(autodiff.diff_to_code(logf, [v1, v2], 'gamma_logfdd', 2))
    file.write(autodiff.diff_to_code(logf, [v1, v2], 'gamma_logfddd', 3))

file.close()

# pd, pdd done manually


import numpy as np
import scipy.stats

from . import expon_libs as exp_libs
from . import utils as cp_utils


def ppf(x, p=np.arange(0.1, 1.0, 0.1), means=False, waicscores=False, 
           logscores=False, rust=False, nrust=100000, debug=False):
    
    """
    Exponential Distribution Predictions Based on a Calibrating Prior

    Parameters
    ----------
    x : array_like
        Observed data, must be non-negative and finite.
    p : array_like, optional
        Probabilities at which to compute quantiles (default: np.arange(0.1, 1.0, 0.1)).
    means : bool, optional
        Whether to compute means (default: False).
    waicscores : bool, optional
        Whether to compute WAIC scores (default: False).
    logscores : bool, optional
        Whether to compute log scores (default: False).
    rust : bool, optional
        Whether to use the "rust" simulation method (default: False).
    nrust : int, optional
        Number of simulations for "rust" (default: 100000).
    debug : bool, optional
        If True, print debug information (default: False).

    Returns
    -------
    dict
        Dictionary containing ML and calibrating prior quantiles, means, scores, and method information.

    Notes
    -----
    The exponential distribution has exceedance distribution function
    S(x;λ)=exp(-λ x), x ≥ 0, λ > 0.
    The calibrating prior is the right Haar prior: π(λ) ∝ 1/λ,
    as given in Jewson et al. (2025).
    """

    x = cp_utils.to_array(x)
    p = cp_utils.to_array(p)
    
    # 1 intro
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NaN"
    assert np.all(np.isfinite(p)) and not np.any(np.isnan(p)), "p must be finite and not NaN"
    assert np.all(p > 0) and np.all(p < 1), "p must be between 0 and 1"
    assert not np.any(x < 0), "x must be non-negative"
    
    alpha = 1 - p
    nx = len(x)
    sumx = np.sum(x)
    
    # 2 ml param estimate
    v1hat = nx / sumx
    ml_params = v1hat
    
    # 3 aic
    ml_value = np.sum(scipy.stats.expon.logpdf(x, scale=1/v1hat))
    maic = cp_utils.make_maic(ml_value, nparams=1)
    
    # 4 ml quantiles (vectorized over alpha)
    ml_quantiles = scipy.stats.expon.ppf(1-alpha, scale=1/v1hat)
    ldd = "only relevant for DMGS models, not analytic models"
    lddi = "only relevant for DMGS models, not analytic models"
    expinfmat = nx / (v1hat * v1hat)
    expinfmati = 1 / expinfmat
    standard_errors = np.sqrt(expinfmati)
    # if(extras){
    #     expinfmat=nx/(v1hat*v1hat)
    #     expinfmati=1/expinfmat
    #     standard_errors=sqrt(expinfmati)
    # }else{
    #     expinfmat="extras not selected"
    #     expinfmati="extras not selected"
    #     standard_errors="extras not selected"
    # }
    
    # 5 rh_quantiles (vectorized over alpha)
    rh_quantiles = sumx * ((alpha**(-1/nx)) - 1)
    
    # 6 means (might as well always calculate)
    ml_mean = 1 / v1hat
    rh_mean = sumx / (nx - 1)
    
    # 7 waicscores
    waic = exp_libs.exp_waic(waicscores, x, v1hat)
    waic1 = waic['waic1']
    waic2 = waic['waic2']
    
    # 8 logscores
    logscores_result = exp_libs.exp_logscores(logscores, x)
    ml_oos_logscore = logscores_result['ml_oos_logscore']
    rh_oos_logscore = logscores_result['rh_oos_logscore']
    
    # 9 rust
    ru_quantiles = "rust not selected"
    if rust:
        rustsim = rvs(nrust, x, rust=True, mlcp=False)
        ru_quantiles = exp_libs.makeq(rustsim['ru_deviates'], p)
    
    # return
    return {
        'ml_params': ml_params,
        'ml_value': ml_value,
        'standard_errors': standard_errors,
        'ml_quantiles': ml_quantiles,
        'cp_quantiles': rh_quantiles,
        'ru_quantiles': ru_quantiles,
        'maic': maic,
        'waic1': waic1,
        'waic2': waic2,
        'ml_oos_logscore': ml_oos_logscore,
        'cp_oos_logscore': rh_oos_logscore,
        'ml_mean': ml_mean,
        'cp_mean': rh_mean,
        'cp_method': cp_utils.analytic_cpmethod()
    }

def rvs(n, x, rust=False, mlcp=True, debug=False):
    """
    Random number generation for exponential calibrating prior

    Parameters
    ----------
    n : int
        Number of random variates to generate.
    x : array_like
        Observed data, must be non-negative and finite.
    rust : bool, optional
        Whether to use the "rust" simulation method (default: False).
    mlcp : bool, optional
        Whether to use ML and calibrating prior (default: True).
    debug : bool, optional
        If True, print debug information (default: False).

    Returns
    -------
    dict
        Dictionary containing ML, calibrating prior, and rust deviates, parameters, and method information.
    """

    x = cp_utils.to_array(x)
    
    # stopifnot(is.finite(n),!is.na(n),is.finite(x),!is.na(x),!x<0)
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NaN"
    assert not np.any(x < 0), "x must be non-negative"
    
    ml_params = "mlcp not selected"
    ml_deviates = "mlcp not selected"
    cp_deviates = "mlcp not selected"
    ru_deviates = "rust not selected"
    
    if mlcp:
        q = ppf(x, np.random.uniform(size=n))
        ml_params = q['ml_params']
        ml_deviates = q['ml_quantiles']
        cp_deviates = q['cp_quantiles']
    
    if rust:
        th = texp_cp(n, x)['theta_samples']
        ru_deviates = np.zeros(n)
        for i in range(n):
            ru_deviates[i] = np.random.exponential(scale=1/th[i])
    
    op = {
        'ml_params': ml_params,
        'ml_deviates': ml_deviates,
        'cp_deviates': cp_deviates,
        'ru_deviates': ru_deviates,
        'cp_method': cp_utils.analytic_cpmethod()
    }
    
    return op

def pdf(x, y=None, rust=False, nrust=1000, debug=False):
    """
    Density function for exponential calibrating prior

    Parameters
    ----------
    x : array_like
        Observed data, must be non-negative and finite.
    y : array_like, optional
        Points at which to evaluate the density (default: x).
    rust : bool, optional
        Whether to use the "rust" simulation method (default: False).
    nrust : int, optional
        Number of simulations for "rust" (default: 1000).
    debug : bool, optional
        If True, print debug information (default: False).

    Returns
    -------
    dict
        Dictionary containing ML, calibrating prior, and rust densities, parameters, and method information.
    """
    x = cp_utils.to_array(x)
    if y is None:
        y = x
    else:
        y = cp_utils.to_array(y)
        
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NaN"
    assert np.all(np.isfinite(y)) and not np.any(np.isnan(y)), "y must be finite and not NaN"
    assert not np.any(x < 0), "x must be non-negative"
    assert not np.any(y < 0), "y must be non-negative"
    
    dd = exp_libs.dexpsub(x=x, y=y)
    ru_pdf = "rust not selected"
    if rust:
        th = texp_cp(nrust, x)['theta_samples']
        ru_pdf = np.zeros(len(y))
        for ir in range(nrust):
            ru_pdf += scipy.stats.expon.pdf(y, scale=1/th[ir])
        ru_pdf = ru_pdf / nrust
    
    op = {
        'ml_params': dd['ml_params'],
        'ml_pdf': dd['ml_pdf'],
        'cp_pdf': dd['rh_pdf'],
        'ru_pdf': ru_pdf,
        'cp_method': cp_utils.analytic_cpmethod()
    }
    return op

def cdf(x, y=None, rust=False, nrust=1000, debug=False):
    """
    Cumulative distribution function for exponential calibrating prior

    Parameters
    ----------
    x : array_like
        Observed data, must be non-negative and finite.
    y : array_like, optional
        Points at which to evaluate the CDF (default: x).
    rust : bool, optional
        Whether to use the "rust" simulation method (default: False).
    nrust : int, optional
        Number of simulations for "rust" (default: 1000).
    debug : bool, optional
        If True, print debug information (default: False).

    Returns
    -------
    dict
        Dictionary containing ML, calibrating prior, and rust CDFs, parameters, and method information.
    """

    x = cp_utils.to_array(x)
    if y is None:
        y = x
    else:
        y = cp_utils.to_array(y)
        
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NaN"
    assert np.all(np.isfinite(y)) and not np.any(np.isnan(y)), "y must be finite and not NaN"
    assert not np.any(x < 0), "x must be non-negative"
    assert not np.any(y < 0), "y must be non-negative"
    
    dd = exp_libs.dexpsub(x=x, y=y)
    ru_cdf = "rust not selected"
    if rust:
        th = texp_cp(nrust, x)['theta_samples']
        ru_cdf = np.zeros(len(y))
        for ir in range(nrust):
            ru_cdf += scipy.stats.expon.cdf(y, scale=1/th[ir])
        ru_cdf = ru_cdf / nrust
    
    op = {
        'ml_params': dd['ml_params'],
        'ml_cdf': dd['ml_cdf'],
        'cp_cdf': dd['rh_cdf'],
        'ru_cdf': ru_cdf,
        'cp_method': cp_utils.analytic_cpmethod()
    }
    return op

def texp_cp(n, x, debug=False):
    """
    Not implemented: Theta sampling for exponential calibrating prior

    Parameters
    ----------
    n : int
        Number of samples to generate.
    x : array_like
        Observed data, must be non-negative and finite.
    debug : bool, optional
        If True, print debug information (default: False).

    Returns
    -------
    dict
        Dictionary containing theta samples.
    """
    raise Exception('texp and rust are not yet implemented; please use DMGS.')
    # stopifnot(is.finite(n),!is.na(n),is.finite(x),!is.na(x),!x<0)
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NaN"
    assert not np.any(x < 0), "x must be non-negative"
    
    t = ru(exp_libs.exp_logf, x=x, n=n, d=1, init=np.mean(x))
    
    return {'theta_samples': t['sim_vals']}




import numpy as np
from scipy.optimize import minimize
from scipy import stats
import warnings

from . import utils as cp_utils
from . import evaluate_dmgs_equation as cp_dmgs
from . import genextreme_p1_libs as cp_gev_p1_b
from . import genextreme_p1_derivs as cp_gev_p1_c


def ppf(x, t, t0=None, n0=None, p=np.arange(0.1, 1.0, 0.1), ics=np.array([0,0,0,0]),
               fdalpha=0.01, minxi=-1, maxxi=1, means=False, waicscores=False, extramodels=False,
               pdf=False, dmgs=True, rust=False, nrust=100000, predictordata=True,
               centering=True, debug=False):
    """Generalized Extreme Value Distribution with a Predictor, Predictions Based on a Calibrating Prior.

    Parameters
    ----------
    x : array_like
        Input data array.
    t : array_like
        Predictor variable array.
    t0 : array_like or None, optional
        Predictor values for prediction (default is None).
    n0 : int or None, optional
        Number of prediction points (default is None).
    p : array_like, optional
        Probabilities for quantile calculation (default is np.arange(0.1, 1.0, 0.1)).
    ics : array_like, optional
        Initial parameter estimates for optimization (default is [0, 0, 0, 0]).
    fdalpha : float, optional
        Finite difference step for PDF estimation (default is 0.01).
    minxi : float, optional
        Minimum value for the shape parameter xi (default is -1).
    maxxi : float, optional
        Maximum value for the shape parameter xi (default is 1).
    means : bool, optional
        Whether to compute means for extra models (default is False).
    waicscores : bool, optional
        Whether to compute WAIC scores (default is False).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    pdf : bool, optional
        Whether to compute PDFs (default is False).
    dmgs : bool, optional
        Whether to use the DMGS method (default is True).
    rust : bool, optional
        Whether to use the rust method (default is False).
    nrust : int, optional
        Number of rust simulations (default is 100000).
    predictordata : bool, optional
        Whether to return predictor data (default is True).
    centering : bool, optional
        Whether to center the predictor variable (default is True).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing ML parameters, quantiles, PDFs, means, WAIC scores, and other results.
    """

    x = cp_utils.to_array(x)
    t = cp_utils.to_array(t)
    
    # Input validation
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x))
    assert np.all(np.isfinite(p)) and not np.any(np.isnan(p)) and np.all(p > 0) and np.all(p < 1)
    assert len(ics) == 4

    # 1 intro
    alpha = 1 - p
    nx = len(x)
    nalpha = len(alpha)
    t0 = cp_utils.maket0(t0, n0, t)
    
    if debug:
        print(f" t0={t0}")
    
    if pdf:
        dalpha = np.minimum(fdalpha * alpha, fdalpha * (1 - alpha))
        alpham = alpha - dalpha
        alphap = alpha + dalpha

    # 2 centering
    if centering:
        meant = np.mean(t)
        t = t - meant
        t0 = t0 - meant

    # 3 ml param estimate
    if debug:
        print(" ml param estimate")
    
    ics = cp_gev_p1_b.gev_p1_setics(x, t, ics)
    
    # Define objective function for optimization
    def neg_loglik(params):
        return -cp_gev_p1_b.gev_p1_loglik(params, x, t)
    
    opt1 = minimize(neg_loglik, ics, method='BFGS')
    if not opt1.success:
        warnings.warn('scipy.optimize.minimize fails due to precision loss.')
    v1hat = opt1.x[0]
    v2hat = opt1.x[1] 
    v3hat = opt1.x[2]
    v4hat = opt1.x[3]
    ml_params = np.array([v1hat, v2hat, v3hat, v4hat])
    
    if debug:
        print(f" ml_params={ml_params}")
    
    if abs(v4hat) >= 1:
        revert2ml = True
    else:
        revert2ml = False

    # 4 predictordata
    prd = cp_gev_p1_b.gev_p1_predictordata(predictordata, x, t, t0, ml_params)
    predictedparameter = prd['predictedparameter']
    adjustedx = prd['adjustedx']

    # 5 aic
    if debug:
        print(" aic")
    
    ml_value = -opt1.fun
    maic = cp_utils.make_maic(ml_value, nparams=4)

    # 6 calc ml quantiles and density
    if debug:
        print(" ml_quantiles")
    
    ml_quantiles = cp_gev_p1_b.qgev_p1((1-alpha), t0, ymn=v1hat, slope=v2hat, sigma=v3hat, xi=v4hat)
    
    if v4hat < 0:
        ml_max = (v1hat + v2hat * t0) - v3hat / v4hat
    else:
        ml_max = np.inf
    
    fhat = cp_gev_p1_b.dgev_p1(ml_quantiles, t0, ymn=v1hat, slope=v2hat, sigma=v3hat, xi=v4hat, log=False)
    
    if debug:
        print(f" ml_quantiles={ml_quantiles}")
        print(f" fhat={fhat}")

    # dmgs initialization
    standard_errors = "dmgs not selected"
    rh_flat_quantiles = "dmgs not selected"
    cp_quantiles = "dmgs not selected"
    ru_quantiles = "dmgs not selected"
    rh_flat_pdf = "dmgs not selected"
    ml_pdf = "dmgs not selected"
    cp_pdf = "dmgs not selected"
    waic1 = "dmgs not selected"
    waic2 = "dmgs not selected"
    ml_mean = "dmgs not selected"
    cp_mean = "dmgs not selected"
    rh_flat_mean = "dmgs not selected"
    cp_method = "dmgs not selected"
    
    if dmgs and not revert2ml:
        # 7 alpha pdf stuff
        # -for now, I only make the pdf. I could add cdf too I suppose. Somehow.
        if pdf:
            ml_quantilesm = cp_gev_p1_b.qgev_p1((1-alpham), t0, ymn=v1hat, slope=v2hat, sigma=v3hat, xi=v4hat)
            ml_quantilesp = cp_gev_p1_b.qgev_p1((1-alphap), t0, ymn=v1hat, slope=v2hat, sigma=v3hat, xi=v4hat)
            fhatm = cp_gev_p1_b.dgev_p1(ml_quantilesm, t0, ymn=v1hat, slope=v2hat, sigma=v3hat, xi=v4hat, log=False)
            fhatp = cp_gev_p1_b.dgev_p1(ml_quantilesp, t0, ymn=v1hat, slope=v2hat, sigma=v3hat, xi=v4hat, log=False)

        # 8 ldd
        ldd = cp_gev_p1_c.gev_p1_ldda(x, t, v1hat, v2hat, v3hat, v4hat)
        
        if debug:
            print(f" ldd={ldd}")
            print(f" det(ldd)={np.linalg.det(ldd)}")
        
        lddi = np.linalg.solve(ldd, np.eye(ldd.shape[0]))
        standard_errors = cp_utils.make_se(nx, lddi)

        # 10 calculate lddd
        if debug:
            print(" lddd")
        
        lddd = cp_gev_p1_c.gev_p1_lddda(x, t, v1hat, v2hat, v3hat, v4hat)

        # 11 mu1
        if debug:
            print(" calculate mu1")
        
        mu1 = cp_gev_p1_c.gev_p1_mu1fa(alpha, t0, v1hat, v2hat, v3hat, v4hat)

        if pdf:
            mu1m = cp_gev_p1_c.gev_p1_mu1fa(alpham, t0, v1hat, v2hat, v3hat, v4hat)
            mu1p = cp_gev_p1_c.gev_p1_mu1fa(alphap, t0, v1hat, v2hat, v3hat, v4hat)

        # 12 mu2
        if debug:
            print(" calculate mu2")
        
        mu2 = cp_gev_p1_c.gev_p1_mu2fa(alpha, t0, v1hat, v2hat, v3hat, v4hat)

        if pdf:
            mu2m = cp_gev_p1_c.gev_p1_mu2fa(alpham, t0, v1hat, v2hat, v3hat, v4hat)
            mu2p = cp_gev_p1_c.gev_p1_mu2fa(alphap, t0, v1hat, v2hat, v3hat, v4hat)

        # 15 model 4: rh_Flat with flat prior on shape (needs to use 4d version of Bayesian code)
        lambdad_rh_flat = np.array([0, 0, -1/v3hat, 0])
        dq = cp_dmgs.dmgs(lddi, lddd, mu1, lambdad_rh_flat, mu2, dim=4)
        rh_flat_quantiles = ml_quantiles + dq / (nx * fhat)
        
        if pdf:
            dqm = cp_dmgs.dmgs(lddi, lddd, mu1m, lambdad_rh_flat, mu2m, dim=4)
            dqp = cp_dmgs.dmgs(lddi, lddd, mu1p, lambdad_rh_flat, mu2p, dim=4)
            quantilesm = ml_quantilesm + dqm / (nx * fhatm)
            quantilesp = ml_quantilesp + dqp / (nx * fhatp)
            ml_pdf = fhat
            rh_flat_pdf = -(alphap - alpham) / (quantilesp - quantilesm)
        else:
            ml_pdf = fhat
            rh_flat_pdf = "pdf not selected"

        # 17 means
        means_result = cp_gev_p1_b.gev_p1_means(means, t0, ml_params, lddi, lddd,
                                               lambdad_rh_flat, nx, dim=4)
        ml_mean = means_result['ml_mean']
        rh_flat_mean = means_result['rh_flat_mean']

        # 18 waicscores
        waic = cp_gev_p1_b.gev_p1_waic(waicscores, x, t, v1hat, v2hat, v3hat, v4hat,
                                      lddi, lddd, lambdad_rh_flat)
        waic1 = waic['waic1']
        waic2 = waic['waic2']

        # 20 rust
        ru_quantiles = "rust not selected"
        if rust:
            rustsim = rvs(nrust, x, t=t, t0=t0, rust=True, mlcp=False)
            ru_quantiles = cp_utils.makeq(rustsim['ru_deviates'], p)
    else:
        rh_flat_quantiles = ml_quantiles
        ru_quantiles = ml_quantiles
        rh_flat_pdf = ml_pdf
        rh_flat_mean = ml_mean

    # 21 decentering
    if centering:
        ml_params[0] = ml_params[0] - ml_params[1] * meant
        if predictordata:
            predictedparameter = predictedparameter - ml_params[1] * meant

    return {
        'ml_params': ml_params,
        'ml_value': ml_value,
        'predictedparameter': predictedparameter,
        'adjustedx': adjustedx,
        'standard_errors': standard_errors,
        'revert2ml': revert2ml,
        'ml_quantiles': ml_quantiles,
        'ml_max': ml_max,
        'cp_quantiles': rh_flat_quantiles,
        'ru_quantiles': ru_quantiles,
        'ml_pdf': ml_pdf,
        'cp_pdf': rh_flat_pdf,
        'maic': maic,
        'waic1': waic1,
        'waic2': waic2,
        'ml_mean': ml_mean,
        'cp_mean': rh_flat_mean,
        'cp_method': cp_utils.crhpflat_dmgs_cpmethod()
    }

def rvs(n, x, t, t0=None, n0=None, ics=np.array([0,0,0,0]),
               minxi=-1, maxxi=1, extramodels=False, rust=False, mlcp=True, debug=False):
    """
    Random generation for GEV distribution with predictor and calibrating prior.

    Parameters
    ----------
    n : int
        Number of samples to generate.
    x : array_like
        Input data array.
    t : array_like
        Predictor variable array.
    t0 : array_like or None, optional
        Predictor values for prediction (default is None).
    n0 : int or None, optional
        Number of prediction points (default is None).
    ics : array_like, optional
        Initial parameter estimates for optimization (default is [0, 0, 0, 0]).
    minxi : float, optional
        Minimum value for the shape parameter xi (default is -1).
    maxxi : float, optional
        Maximum value for the shape parameter xi (default is 1).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    rust : bool, optional
        Whether to use the rust method (default is False).
    mlcp : bool, optional
        Whether to use ML and CP deviates (default is True).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing ML parameters, ML deviates, CP deviates, RU deviates, and method info.
    """
    
    x = cp_utils.to_array(x)
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x))
    assert np.all(np.isfinite(t)) and not np.any(np.isnan(t))
    assert len(ics) == 4

    t0 = cp_utils.maket0(t0=t0, n0=n0, t=t)

    # centering
    meant = np.mean(t)
    t = t - meant
    t0 = t0 - meant

    ml_params = "mlcp not selected"
    ml_deviates = "mlcp not selected"
    ru_deviates = "rust not selected"
    cp_deviates = "rust not selected"

    if mlcp:
        q = ppf(x, t=t, t0=t0, n0=None, p=np.random.uniform(0, 1, n), ics=ics, extramodels=extramodels)
        ml_params = q['ml_params']
        ml_deviates = q['ml_quantiles']
        ru_deviates = q['ru_quantiles']
        cp_deviates = q['cp_quantiles']

    if rust:
        th = tgev_p1_cp(n, x, t)['theta_samples']
        ru_deviates = np.zeros(n)
        for i in range(n):
            mu = th[i,0] + t0 * th[i,1]
            ru_deviates[i] = stats.genextreme.rvs(c=-th[i,3], loc=mu, scale=th[i,2], size=1)

    # decentering
    if mlcp:
        ml_params[0] = ml_params[0] - ml_params[1] * meant

    op = {
        'ml_params': ml_params,
        'ml_deviates': ml_deviates,
        'cp_deviates': cp_deviates,
        'ru_deviates': ru_deviates,
        'cp_method': cp_utils.crhpflat_dmgs_cpmethod()
    }

    return op

def pdf(x, t, t0=None, n0=None, y=None, ics=np.array([0,0,0,0]),
               minxi=-1, maxxi=1, extramodels=False,
               rust=False, nrust=1000, centering=True, debug=False):
    """
    Density function for GEV distribution with predictor and calibrating prior
    Parameters
    ----------
    x : array_like
        Input data array.
    t : array_like
        Predictor variable array.
    t0 : array_like or None, optional
        Predictor values for prediction (default is None).
    n0 : int or None, optional
        Number of prediction points (default is None).
    y : array_like or None, optional
        Points at which to evaluate the density (default is x).
    ics : array_like, optional
        Initial parameter estimates for optimization (default is [0, 0, 0, 0]).
    minxi : float, optional
        Minimum value for the shape parameter xi (default is -1).
    maxxi : float, optional
        Maximum value for the shape parameter xi (default is 1).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    rust : bool, optional
        Whether to use the rust method (default is False).
    nrust : int, optional
        Number of rust simulations (default is 1000).
    centering : bool, optional
        Whether to center the predictor variable (default is True).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing ML parameters, ML PDF, RU PDF, and method info.
    """
    
    if y is None:
        y = x
        
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x))
    assert np.all(np.isfinite(y)) and not np.any(np.isnan(y))
    assert np.all(np.isfinite(t)) and not np.any(np.isnan(t))
    assert len(ics) == 4

    t0 = cp_utils.maket0(t0, n0, t)

    # centering
    if centering:
        meant = np.mean(t)
        t = t - meant
        t0 = t0 - meant

    ics = cp_gev_p1_b.gev_p1_setics(x, t, ics)
    
    # Define objective function for optimization
    def neg_loglik(params):
        return -cp_gev_p1_b.gev_p1_loglik(params, x, t)
    
    opt1 = minimize(neg_loglik, ics, method='BFGS')
    v1hat = opt1.x[0]
    v2hat = opt1.x[1] 
    v3hat = opt1.x[2]
    v4hat = opt1.x[3]
    
    if v4hat <= (-1):
        revert2ml = True
    else:
        revert2ml = False
        
    ml_params = np.array([v1hat, v2hat, v3hat, v4hat])
    
    dd = cp_gev_p1_b.dgev_p1sub(x=x, t=t, y=y, t0=t0, ics=ics, minxi=minxi, maxxi=maxxi, extramodels=extramodels)
    ru_pdf = "rust not selected"
    ml_params = dd['ml_params']

    if rust and not revert2ml:
        th = tgev_p1_cp(nrust, x, t)['theta_samples']
        ru_pdf = np.zeros(len(y))
        for ir in range(nrust):
            mu = th[ir,0] + t0 * th[ir,1]
            ru_pdf = ru_pdf + stats.genextreme.pdf(y, c=-th[ir,3], loc=mu, scale=th[ir,2])
        ru_pdf = ru_pdf / nrust
    else:
        ru_pdf = dd['ml_pdf']

    # decentering
    if centering:
        ml_params[0] = ml_params[0] - ml_params[1] * meant

    op = {
        'ml_params': ml_params,
        'ml_pdf': dd['ml_pdf'],
        'revert2ml': revert2ml,
        'ru_pdf': ru_pdf,
        'cp_method': cp_utils.nopdfcdfmsg()
    }

    return op

def cdf(x, t, t0=None, n0=None, y=None, ics=np.array([0,0,0,0]),
               minxi=-1, maxxi=1, extramodels=False,
               rust=False, nrust=1000, centering=True, debug=False):
    """
    Distribution function for GEV distribution with predictor and calibrating prior.

    Parameters
    ----------
    x : array_like
        Input data array.
    t : array_like
        Predictor variable array.
    t0 : array_like or None, optional
        Predictor values for prediction (default is None).
    n0 : int or None, optional
        Number of prediction points (default is None).
    y : array_like or None, optional
        Points at which to evaluate the CDF (default is x).
    ics : array_like, optional
        Initial parameter estimates for optimization (default is [0, 0, 0, 0]).
    minxi : float, optional
        Minimum value for the shape parameter xi (default is -1).
    maxxi : float, optional
        Maximum value for the shape parameter xi (default is 1).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    rust : bool, optional
        Whether to use the rust method (default is False).
    nrust : int, optional
        Number of rust simulations (default is 1000).
    centering : bool, optional
        Whether to center the predictor variable (default is True).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing ML parameters, ML CDF, RU CDF, and method info.
    """
    
    if y is None:
        y = x
        
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x))
    assert np.all(np.isfinite(y)) and not np.any(np.isnan(y))
    assert np.all(np.isfinite(t)) and not np.any(np.isnan(t))
    assert len(ics) == 4

    t0 = cp_utils.maket0(t0, n0, t)
    
    # centering
    if centering:
        meant = np.mean(t)
        t = t - meant
        t0 = t0 - meant

    ics = cp_gev_p1_b.gev_p1_setics(x, t, ics)
    
    # Define objective function for optimization
    def neg_loglik(params):
        return -cp_gev_p1_b.gev_p1_loglik(params, x, t)
    
    opt1 = minimize(neg_loglik, ics, method='BFGS')
    v1hat = opt1.x[0]
    v2hat = opt1.x[1] 
    v3hat = opt1.x[2]
    v4hat = opt1.x[3]
    
    if v4hat <= (-1):
        revert2ml = True
    else:
        revert2ml = False
        
    ml_params = np.array([v1hat, v2hat, v3hat, v4hat])
    
    dd = cp_gev_p1_b.dgev_p1sub(x=x, t=t, y=y, t0=t0, ics=ics, minxi=minxi, maxxi=maxxi, extramodels=extramodels)
    ru_cdf = "rust not selected"
    ml_params = dd['ml_params']

    if rust and not revert2ml:
        th = tgev_p1_cp(nrust, x, t)['theta_samples']
        ru_cdf = np.zeros(len(y))
        for ir in range(nrust):
            mu = th[ir,0] + t0 * th[ir,1]
            ru_cdf = ru_cdf + stats.genextreme.cdf(y, c=-th[ir,3], loc=mu, scale=th[ir,2])
        ru_cdf = ru_cdf / nrust
    else:
        ru_pdf = dd['ml_pdf']

    # decentering
    if centering:
        ml_params[0] = ml_params[0] - ml_params[1] * meant

    op = {
        'ml_params': ml_params,
        'ml_cdf': dd['ml_cdf'],
        'revert2ml': revert2ml,
        'ru_cdf': ru_cdf,
        'cp_method': cp_utils.nopdfcdfmsg()
    }
    
    return op

def tgev_p1_cp(n, x, t, ics=np.array([0,0,0,0]),
               extramodels=False, debug=False):
    """Not implemented: Theta sampling for GEV distribution with predictor and calibrating prior

    Parameters
    ----------
    n : int
        Number of theta samples to generate.
    x : array_like
        Input data array.
    t : array_like
        Predictor variable array.
    ics : array_like, optional
        Initial parameter estimates for optimization (default is [0, 0, 0, 0]).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing theta samples.
    """

    raise Exception('tgev_p1_cp (and rust) is not yet implemented in fitdistcp; please use the dmgs method.')
    
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x))
    assert np.all(np.isfinite(t)) and not np.any(np.isnan(t))
    assert len(ics) == 4

    # centering
    meant = np.mean(t)
    t = t - meant

    ics = cp_gev_p1_b.gev_p1_setics(x, t, ics)
    th = cp_utils.ru(cp_gev_p1_b.gev_p1_logf, x=x, t=t, n=n, d=4, init=ics)
    theta_samples = th['sim_vals']
    
    # decentering
    theta_samples[:,0] = theta_samples[:,0] - theta_samples[:,1] * meant

    return {'theta_samples': theta_samples}




import numpy as np
import scipy.stats as stats
from scipy.optimize import minimize
from scipy.special import gamma, digamma, polygamma
import sys

from . import utils as cp_utils
from . import genextreme_p1_derivs as cp_gev_p1_c


def rgev_p1_minmax(nx, mu, sigma, xi, tt, minxi=-0.45, maxxi=0.45, centering=True):
    """rgev for gev_p1 but with maxlik xi within bounds"""
    xihat = -9999
    if centering:
        tt = tt - np.mean(tt)
    
    while (xihat < minxi) or (xihat > maxxi):  # 0.46 also works...0.47 doesn't
        xx = stats.genextreme.rvs(c=-xi, loc=mu, scale=sigma, size=nx)
        ics = gev_p1_setics(xx, tt, np.array([0, 0, 0, 0]))
        
        # Define objective function for minimization (negative log-likelihood)
        def neg_loglik(params):
            return -gev_p1_loglik(params, xx, tt)
        
        opt1 = minimize(neg_loglik, ics, method='BFGS')
        xihat = opt1.x[3]
    
    return xx


def gev_p1_waic(waicscores, x, t0, v1hat, v2hat, v3hat, v4hat, lddi, lddd, lambdad):
    """WAIC calculation """
    if waicscores:
        f1f = cp_gev_p1_c.gev_p1_f1fw(x, t0, v1hat, v2hat, v3hat, v4hat)
        f2f = cp_gev_p1_c.gev_p1_f2fw(x, t0, v1hat, v2hat, v3hat, v4hat)
        fhatx = dgev_p1(x, t0, ymn=v1hat, slope=v2hat, sigma=v3hat, xi=v4hat, log=False)
        waic = cp_utils.make_waic(x, fhatx, lddi, lddd, f1f, lambdad, f2f, dim=4)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
    else:
        waic1 = "waicscores not selected"
        waic2 = "waicscores not selected"
    
    return {'waic1': waic1, 'waic2': waic2}


def gev_p1_predictordata(predictordata, x, t, t0, params):
    """Predicted Parameter and Generalized Residuals"""
    if predictordata:
        # calculate the probabilities of the data using the fitted model
        a = params[0]
        b = params[1]
        sc = params[2]
        sh = params[3]
        mu = a + b * t
        px = stats.genextreme.cdf(x, c=-sh, loc=mu, scale=sc)
        
        # calculate the quantiles for those probabilities at t0
        mu0 = a + b * t0
        qx = stats.genextreme.ppf(px, c=-sh, loc=mu0, scale=sc)
        predictedparameter = mu
        adjustedx = qx
    else:
        predictedparameter = "predictordata not selected"
        adjustedx = "predictordata not selected"
    
    return {'predictedparameter': predictedparameter, 'adjustedx': adjustedx}


def gev_p1_logf(params, x, t):
    """Logf for RUST    """
    a = params[0]
    b = params[1]
    sc = np.maximum(params[2], np.sqrt(sys.float_info.epsilon))
    sh = params[3]
    mu = a + b * t
    logf = np.sum(stats.genextreme.logpdf(x, c=-sh, loc=mu, scale=sc)) - np.log(sc)
    return logf


def gev_p1_setics(x, t, ics):
    """Set initial conditions"""
    nx = len(x)
    if (ics[0] == 0) and (ics[1] == 0) and (ics[2] == 0) and (ics[3] == 0):
        # Linear regression equivalent to R's lm(x~t)
        A = np.vstack([np.ones(len(t)), t]).T
        coeffs = np.linalg.lstsq(A, x, rcond=None)[0]
        ics[0] = coeffs[0]  # intercept
        ics[1] = coeffs[1]  # slope
        xhat = ics[0] + ics[1] * t
        ics[2] = np.sqrt(np.sum((x - xhat)**2) / nx)
        ics[3] = 0
    
    return ics


def gev_p1_loglik(vv, x, t):
    """Observed log-likelihood function  """
    n = len(x)
    mu = vv[0] + vv[1] * t  # so mean is a vector, just like x
    loglik = np.sum(stats.genextreme.logpdf(x, c=-vv[3], loc=mu, 
                                          scale=max(vv[2], sys.float_info.epsilon)))
    return loglik


def gev_p1_checkmle(ml_params, minxi=-1, maxxi=1):
    """Check MLE"""
    # currently not used, because instead I revert2ml
    v1hat = ml_params[0]
    v2hat = ml_params[1]
    v3hat = ml_params[2]
    v4hat = ml_params[3]
    
    if np.isnan(v1hat):
        raise ValueError("v1hat is NaN")
    if np.isnan(v2hat):
        raise ValueError("v2hat is NaN")
    if np.isnan(v3hat):
        raise ValueError("v3hat is NaN")
    if np.isnan(v4hat):
        raise ValueError("v4hat is NaN")
    
    if v4hat < minxi:
        raise ValueError(f"\n***v4hat={v4hat}=> execution halted because maxlik shape parameter <{minxi}***\n")
    if v4hat > maxxi:
        raise ValueError(f"\n***v4hat={v4hat}=> execution halted because maxlik shape parameter >{maxxi}***\n")


def qgev_p1(p, t0, ymn, slope, sigma, xi):
    """GEVD-with-p1: Quantile function """
    return stats.genextreme.ppf(p, c=-xi, loc=(ymn + slope * t0), scale=sigma)

def dgev_p1(x, t0, ymn, slope, sigma, xi, log=False):
    """GEVD-with-p1: Density function
    
    Args:
        x: data vector
        t0: time value
        ymn: mean parameter
        slope: slope parameter
        sigma: scale parameter
        xi: shape parameter
        log: whether to return log density (default False)
    
    Returns:
        Density values
    """
    mu = ymn + slope * t0
    if log:
        return stats.genextreme.logpdf(x, c=-xi, loc=mu, scale=sigma)
    else:
        return stats.genextreme.pdf(x, c=-xi, loc=mu, scale=sigma)


def pgev_p1(y, t0, ymn, slope, sigma, xi):
    """GEVD-with-p1: Distribution function """
    return stats.genextreme.cdf(y, c=-xi, loc=(ymn + slope * t0), scale=sigma)


def gev_p1_means(means, t0, ml_params, lddi, lddd, lambdad_rh_flat, nx, dim=4):
    """Analytical expressions for Predictive Means
    RHP mean based on the expectation of DMGS equation 2.1
    """

    if means:
        # intro
        eulerconstant = 0.57721566490153286060651209008240243104215933593992
        ymn = ml_params[0]
        slope = ml_params[1]
        sigma = ml_params[2]
        xi = ml_params[3]
        
        # set up derivative arrays
        meand1 = np.zeros((4, 1))
        meand2 = np.zeros((4, 4, 1))  # but all zero for gumbel
        
        if ml_params[3] == 0:
            # xi=0 case
            
            # mle
            ml_mean = ymn + slope * t0 + sigma * eulerconstant
            # calculate first derivative array for bayesian xi=0 cases
            meand1[0, 0] = 1
            meand1[1, 0] = t0
            meand1[2, 0] = eulerconstant
            meand1[3, 0] = 0
            # meand2 is all zero as initialized
            
        else:
            # non-gumbel case
            
            g0 = gamma(1 - xi)
            g1 = g0 * digamma(1 - xi)
            g2 = (polygamma(1, 1 - xi) * g0 * g0 + g1 * g1) / g0  # trigamma is polygamma(1, x)
            # mle
            ml_mean = ymn + slope * t0 + sigma * (g0 - 1) / xi
            # calculate first derivative array for bayesian xi!=0 cases
            meand1[0, 0] = 1
            meand1[1, 0] = t0
            meand1[2, 0] = (g0 - 1) / xi
            meand1[3, 0] = (1 - g0 - xi * g1) * sigma / (xi * xi)
            # calculate second derivative array (only has 1 non-zero term!)
            meand2[2, 3, 0] = (1 - g0 - xi * g1) / (xi * xi)
            meand2[3, 2, 0] = meand2[2, 3, 0]
            meand2[3, 3, 0] = sigma * (-2 + 2 * g0 + 2 * xi * g1 + xi * xi * g2) / (xi * xi * xi)
        
        # I've realized now that when I integrate over xi, the mean in Inf
        rh_flat_mean = np.inf
        
    else:
        ml_mean = "means not selected"
        rh_flat_mean = "means not selected"
    
    # return
    return {'ml_mean': ml_mean, 'rh_flat_mean': rh_flat_mean}


def dgev_p1sub(x, t, y, t0, ics, minxi, maxxi, extramodels=False):
    """Densities for 5 predictions """
    nx = len(x)
    
    ics = gev_p1_setics(x, t, ics)
    
    # Define objective function for minimization (negative log-likelihood)
    def neg_loglik(params):
        return -gev_p1_loglik(params, x, t)
    
    opt = minimize(neg_loglik, ics, method='BFGS')
    v1hat = opt.x[0]
    v2hat = opt.x[1]
    v3hat = opt.x[2]
    v4hat = min(maxxi, max(minxi, opt.x[3]))
    ml_params = np.array([v1hat, v2hat, v3hat, v4hat])
    
    # now that I've dropped dmgs, not sure I need this anymore
    # muhat0 = v1hat + v2hat * t0
    # y = fixgevrange(y, muhat0, v3hat, v4hat)
    
    # mle
    ml_pdf = dgev_p1(y, t0, ymn=v1hat, slope=v2hat, sigma=v3hat, xi=v4hat)
    ml_cdf = pgev_p1(y, t0, ymn=v1hat, slope=v2hat, sigma=v3hat, xi=v4hat)
    
    # return
    return {
        'ml_params': ml_params,
        'ml_pdf': ml_pdf,
        'ml_cdf': ml_cdf
    }

